from ._cornerpoint_grid import (
    CornerpointGrid,
    InvalidEgridFileError,
    MapAxes,
    InvalidGridError,
)
from ._summary_reader import SummaryReader, InvalidSummaryError, SummaryKeyword
from ._summary_keys import (
    SummaryKeyType,
    history_key,
    is_rate,
    make_summary_key,
    InvalidSummaryKeyError,
)

__all__ = [
    "CornerpointGrid",
    "InvalidEgridFileError",
    "MapAxes",
    "InvalidGridError",
    "SummaryReader",
    "SummaryKeyword",
    "InvalidSummaryError",
    "SummaryKeyType",
    "history_key",
    "is_rate",
    "make_summary_key",
    "InvalidSummaryKeyError",
]



from hypothesis import HealthCheck, settings


settings.register_profile(
    "no_timeouts",
    deadline=None,
    suppress_health_check=[HealthCheck.too_slow],
    print_blob=True,
)
settings.load_profile("no_timeouts")



from __future__ import annotations
import os
from typing import Self, Any, IO, TypeVar, Callable
from dataclasses import dataclass
from numpy import typing as npt
import numpy as np
import resfo
import scipy.optimize
import warnings
import heapq
from functools import cached_property


class InvalidEgridFileError(ValueError):
    pass


class InvalidGridError(ValueError):
    pass


@dataclass
class MapAxes:
    """The axes of the map coordinate system.

    Usually, a corner-point grid contains x,y values that needs to be transformed
    into a map coordinate system (which could be :term:`UTM-coordinates`). That
    coordinate system is represented by MapAxes.

    Note that regardless of the size of the axes, when transforming from the grid
    coordinate system to the map coordinate system, scaling is not applied.

    Attributes:
        y_axis:
            A point along the map y axis.
        origin:
            The origin of the map coordinate system.
        x_axis:
            A point along the map x axis.
    """

    y_axis: tuple[np.float32, np.float32]
    origin: tuple[np.float32, np.float32]
    x_axis: tuple[np.float32, np.float32]

    def transform_map_points(
        self, points: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        """Transforms points from map coordinates to grid coordinates.

        Scaling according to length of the axes is not applied.

        Returns:
            The given map points in the grid coordinate system.
        """
        translated = points - np.array([*self.origin, 0])
        tx = translated[:, 0]
        ty = translated[:, 1]
        x_vec = (self.x_axis[0] - self.origin[0], self.x_axis[1] - self.origin[1])
        y_vec = (self.y_axis[0] - self.origin[0], self.y_axis[1] - self.origin[1])
        x_norm = np.sqrt(x_vec[0] ** 2 + x_vec[1] ** 2)
        x_unit = (x_vec[0] / x_norm, x_vec[1] / x_norm)
        y_norm = np.sqrt(y_vec[0] ** 2 + y_vec[1] ** 2)
        y_unit = (y_vec[0] / y_norm, y_vec[1] / y_norm)
        norm = 1.0 / (x_unit[0] * y_unit[1] - x_unit[1] * y_unit[0])
        return np.column_stack(
            [
                (tx * y_unit[1] - ty * y_unit[0]) * norm,
                (-tx * x_unit[1] + ty * x_unit[0]) * norm,
                translated[:, 2],
            ]
        )


@dataclass
class CornerpointGrid:
    """
    A :term:`corner-point grid` is a tessellation of a 3D volume where
    each cell is a hexahedron.

    Each cell is identified by a integer coordinate (i,j,k).
    For each i,j there is are four straight lines, defined by their end-points
    called a :term:`pillar`. The end-points form two surfaces, one
    for the top end-points and one for the bottom end points, which
    are in the CornerpointGrid.coord array.

    For the cell at position i,j,k, its eight corner vertices are defined by
    giving the z values along the pillars at [(i,j), (i+1, j), (i, j+1), (i+1, j+1)]
    which are in the CornerpointGrid.zcorn array.


    Attributes:
        coord:
            A (ni+1, nj+1, 2, 3) array where coord[i,j,0] is the top end point
            of the i,j pillar and coord[i,j,1] is the corresponding bottom end point.
        zcorn:
            A (ni, nj, nk, 8) array where zcorn[i,j,k] is the z value of
            the 8 corners of the cell at i,j,k. The order of the corner z values
            are as follows:
            [TSW, TSE, TNW, TNE, BSW, BSE, BNW, BNE] where N(orth) means higher y,
            E(east) means higher x, T(op) means lower z (when z is interpreted as depth).

        map_axes:
            Optionally each point is interpreted to be relative to some map
            coordinate system. Defaults to the unit coordinate system with
            origin at (0,0).

    """

    coord: npt.NDArray[np.float32]
    zcorn: npt.NDArray[np.float32]
    map_axes: MapAxes | None = None

    def __post_init__(self) -> None:
        if len(self.coord.shape) != 4 or self.coord.shape[2:4] != (2, 3):
            raise InvalidGridError(f"coord had invalid dimensions {self.coord.shape}")
        if len(self.zcorn.shape) != 4 or self.zcorn.shape[-1] != 8:
            raise InvalidGridError(f"zcorn had invalid dimensions {self.zcorn.shape}")
        ni = self.coord.shape[0] - 1
        nj = self.coord.shape[1] - 1
        if self.zcorn.shape[0] != ni or self.zcorn.shape[1] != nj:
            raise InvalidGridError(
                "zcorn and coord dimensions do not match:"
                f" {self.zcorn.shape} vs {self.coord.shape}"
            )

    @classmethod
    def read_egrid(cls, file_like: str | os.PathLike[str] | IO[Any]) -> Self:
        """Read the global grid from an .EGRID or .FEGRID file.

        If the EGRID contains Local Grid Refinements or Coarsening Groups,
        that is silently ignored and only the host grid is read. Radial grids
        are not supported and will cause InvalidEgridFileError to be raised.

        Args:
            file_like:
                The EGRID file, could either be a filename, pathlike or an opened
                EGRID file. The function also handles formatted egrid files (.FEGRID).
                Whether the file is formatted or not is determined by looking at the
                extension a filepath is given and by whether the stream is a byte-stream
                (unformatted) or a text-stream when an opened file is given.
        Raises:
            InvalidEgridFileError:
                When the egrid file is not valid, or contains a radial grid.
            OSError:
                If the given filepath cannot be opened.

        """
        coord = None
        dims = None
        zcorn = None
        opened = False
        stream = None
        map_axes = None

        try:
            if isinstance(file_like, str):
                filename = file_like
                mode = "rt" if filename.lower().endswith("fegrid") else "rb"
                stream = open(filename, mode=mode)
                opened = True
            elif isinstance(file_like, os.PathLike):
                filename = str(file_like)
                mode = "rt" if filename.lower().endswith("fegrid") else "rb"
                stream = open(filename, mode=mode)
                opened = True
            else:
                filename = getattr(file_like, "name", "unknown stream")
                stream = file_like

            T = TypeVar("T", bound=np.generic)

            def validate_array(
                name: str,
                array: npt.NDArray[T] | resfo.MESS,
                min_length: int | None = None,
            ) -> npt.NDArray[T]:
                if array is resfo.MESS or isinstance(array, resfo.MESS):
                    raise InvalidEgridFileError(
                        f"Expected Array for keyword {name} in {filename} but got MESS"
                    )
                if min_length is not None and len(array) < min_length:
                    raise InvalidEgridFileError(
                        f"{name} in EGRID file {filename} contained too few elements"
                    )

                return array

            def optional_get(array: npt.NDArray[T] | None, index: int) -> T | None:
                if array is None:
                    return None
                if len(array) <= index:
                    return None
                return array[index]

            for entry in resfo.lazy_read(stream):
                kw = entry.read_keyword()
                match kw:
                    case "ZCORN   ":
                        zcorn = validate_array(kw, entry.read_array())
                    case "COORD   ":
                        coord = validate_array(kw, entry.read_array())
                    case "GRIDHEAD":
                        array = validate_array(kw, entry.read_array(), 4)
                        if (reference_number := optional_get(array, 4)) != 0:
                            warnings.warn(
                                f"The global grid in {filename} had "
                                f"reference number {reference_number}, expected 0."
                                " This could indicate that the grid being read"
                                " is actually an LGR grid."
                            )
                        if optional_get(array, 26) not in {0, None}:
                            raise InvalidEgridFileError(
                                f"EGRID file {filename} contains a radial grid"
                                " which is not supported by resfo-utilities."
                            )

                        dims = tuple(array[1:4])
                    case "MAPAXES ":
                        array = validate_array(kw, entry.read_array(), 6)
                        map_axes = MapAxes(
                            (array[0], array[1]),
                            (array[2], array[3]),
                            (array[4], array[5]),
                        )
                    case "ENDGRID ":
                        break

            if coord is None:
                raise InvalidEgridFileError(
                    f"EGRID file {filename} did not contain COORD"
                )
            if zcorn is None:
                raise InvalidEgridFileError(
                    f"EGRID file {filename} did not contain ZCORN"
                )
            if dims is None:
                raise InvalidEgridFileError(
                    f"EGRID file {filename} did not contain dimensions"
                )
        except resfo.ResfoParsingError as err:
            raise InvalidEgridFileError(f"Could not parse EGRID file: {err}") from err
        finally:
            if opened and stream is not None:
                stream.close()
        try:
            coord = np.swapaxes(coord.reshape((dims[1] + 1, dims[0] + 1, 2, 3)), 0, 1)
        except ValueError as err:
            raise InvalidEgridFileError(
                f"COORD size {len(coord)} did not match"
                f" grid dimensions {dims} in {filename}"
            ) from err
        try:
            zcorn = zcorn.reshape(2, dims[0], 2, dims[1], 2, dims[2], order="F")
            zcorn = np.moveaxis(zcorn, [1, 3, 5, 4, 2], [0, 1, 2, 3, 4])
            zcorn = zcorn.reshape((dims[0], dims[1], dims[2], 8))
        except ValueError as err:
            raise InvalidEgridFileError(
                f"ZCORN size {len(zcorn)} did not match"
                f" grid dimensions {dims} in {filename}"
            ) from err
        return cls(coord, zcorn, map_axes)

    def find_cell_containing_point(
        self,
        points: npt.ArrayLike,
        map_coordinates: bool = True,
        tolerance: float = 1.0e-6,
    ) -> list[tuple[int, int, int] | None]:
        """Find a cell in the grid which contains the given point.

        Args:
            points:
                The points to find cells for.
            map_coordinates:
                Whether points are in the map coordinate system.
                Defaults to True.
            tolerance:
                The maximum distance to the cell boundary a point can have to
                be considered to be contained in the cell.

        Returns:
            list of i,j,k indices for each point (or None if the
            point is not contained in any cell.
        """
        points = np.asarray(points)
        result: list[tuple[int, int, int] | None] = []
        if map_coordinates and self.map_axes is not None:
            points = self.map_axes.transform_map_points(points)

        dims = self.zcorn.shape[0:3]
        top = self._pillars_z_plane_intersection(self.zcorn.min())
        bot = self._pillars_z_plane_intersection(self.zcorn.max())

        # This algorithm will for each point p calculate the mesh surface that
        # is the intersection of the pillars with the plane z=p[2]. Then it searches
        # through the quad with a heuristical search that orders each neighbour by
        # the points manhattan distance to the bounding box.
        found = False
        # The use case that the previous point is close to the
        # next point is very common, so we optimize for that.
        prev_ij = None  # The i,j index the previous point was found at

        @dataclass
        class Quad:
            """The quad at index i,j"""

            i: int
            j: int
            p: npt.NDArray[np.float32]
            i_neighbourhood: int
            j_neighbourhood: int

            @cached_property
            def vertices(self) -> npt.NDArray[np.float32]:
                return np.array(
                    [
                        top[self.i, self.j],
                        top[self.i + 1, self.j],
                        top[self.i + 1, self.j + 1],
                        top[self.i, self.j + 1],
                        bot[self.i, self.j],
                        bot[self.i + 1, self.j],
                        bot[self.i + 1, self.j + 1],
                        bot[self.i, self.j + 1],
                    ],
                    dtype=np.float32,
                )

            @cached_property
            def min_x(self) -> np.float32:
                return self.vertices[:, 0].min()

            @cached_property
            def min_y(self) -> np.float32:
                return self.vertices[:, 1].min()

            @cached_property
            def max_x(self) -> np.float32:
                return self.vertices[:, 0].max()

            @cached_property
            def max_y(self) -> np.float32:
                return self.vertices[:, 1].max()

            @cached_property
            def distance_from_bounds(self) -> np.float32:
                """Manhattan distance from the point to the quad bounding box."""
                x_dist = max(self.min_x - self.p[0], self.p[0] - self.max_x, 0)
                y_dist = max(self.min_y - self.p[1], self.p[1] - self.max_y, 0)
                return x_dist + y_dist

            def __lt__(self, other: object) -> bool:
                """Used to order elements in the search queue.

                The Quads are ordered by distance_from_bounds.
                """
                if not isinstance(other, Quad):
                    return False
                return bool(self.distance_from_bounds < other.distance_from_bounds)

        if dims[0] <= 0 or dims[1] <= 0:
            return [None] * len(points)

        for p in points:
            found = False
            if prev_ij is None:
                queue = [
                    Quad(dims[0] // 2, dims[1] // 2, p, dims[0] // 2, dims[1] // 2)
                ]
            else:
                queue = [Quad(*prev_ij, p, 1, 1)]
            visited = set([(queue[0].i, queue[0].j)])
            while queue:
                node = heapq.heappop(queue)
                i = node.i
                j = node.j

                # If the quad contains the point then search through each k index
                # for that quad
                if node.distance_from_bounds <= 2 * tolerance:
                    for k in range(dims[2]):
                        zcorn = self.zcorn[i, j, k]
                        z = p[2]
                        # Prune by bounding box first then check whether point_in_cell
                        if (
                            zcorn.min() - 2 * tolerance
                            <= z
                            <= zcorn.max() + 2 * tolerance
                            and self.point_in_cell(
                                p, i, j, k, tolerance=tolerance, map_coordinates=False
                            )
                        ):
                            prev_ij = (i, j)
                            result.append((i, j, k))
                            found = True
                            break
                if found:
                    break

                # Add each neighbour to the queue if not visited
                size_i = node.i_neighbourhood
                for di in (-1 * size_i, 0, size_i):
                    ni = np.clip(i + di, 0, dims[0] - 1)
                    size_j = node.j_neighbourhood
                    for dj in (-1 * size_j, 0, size_j):
                        nj = np.clip(j + dj, 0, dims[1] - 1)
                        if (ni, nj) not in visited:
                            heapq.heappush(
                                queue,
                                Quad(
                                    ni,
                                    nj,
                                    p,
                                    max(size_i // 2, 1),
                                    max(size_j // 2, 1),
                                ),
                            )
                            visited.add((ni, nj))
            if not found:
                result.append(None)

        return result

    def cell_corners(self, i: int, j: int, k: int) -> npt.NDArray[np.float32]:
        """Array of coordinates for all corners of the cell at i,j,k

        The order of the corners are the same as in zcorn.
        """
        pillar_vertices = np.concatenate(
            [
                self.coord[i, j, :],
                self.coord[i, j + 1, :],
                self.coord[i + 1, j, :],
                self.coord[i + 1, j + 1, :],
            ]
        )
        top = pillar_vertices[::2][[0, 2, 1, 3]]
        bot = pillar_vertices[1::2][[0, 2, 1, 3]]
        top_z = top[:, 2]
        bot_z = bot[:, 2]

        def twice(a: npt.NDArray[Any]) -> npt.NDArray[Any]:
            return np.concatenate([a, a])

        height_diff = twice(bot_z - top_z)

        if np.any(height_diff == 0):
            raise InvalidGridError(
                f"Grid contains zero height pillars with different for cell {i, j, k}"
            )

        t = (self.zcorn[i, j, k] - twice(top_z)) / height_diff

        result = twice(top) + t[:, np.newaxis] * twice(bot - top)

        if not np.all(np.isfinite(result)):
            raise InvalidGridError(
                f"The corners of the cell at {i, j, k} is not well defined"
            )

        return result

    def point_in_cell(
        self,
        points: npt.ArrayLike,
        i: int,
        j: int,
        k: int,
        tolerance: float = 1e-6,
        map_coordinates: bool = True,
    ) -> npt.NDArray[np.bool_]:
        """Whether the points (x,y,z) is in the cell at (i,j,k).

        For containment the cell are considered to have bilinear faces.

        Param:
            points:
                x,y,z triple or array of x,y,z triples to be tested for containment.
            tolerance:
                The tolerance used for numerical precision in the linear
                interpolation calculation.
            map_coordinates:
                Whether the given points are in the mapaxes coordinate system,
                defaults to true.

        Returns:
            Array of boolean values for each triplet describing whether
            it is contained in the cell.
        """
        points = np.asarray(points)
        if len(points.shape) == 1:
            points = points[np.newaxis, :]
        if map_coordinates and self.map_axes is not None:
            points = self.map_axes.transform_map_points(points)

        vertices = self.cell_corners(i, j, k)

        corner_signs = np.array(
            [
                [-1, -1, -1],
                [1, -1, -1],
                [-1, 1, -1],
                [1, 1, -1],
                [-1, -1, 1],
                [1, -1, 1],
                [-1, 1, 1],
                [1, 1, 1],
            ]
        )

        def residual(
            point: tuple[float, float, float],
        ) -> Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]:
            def inner(
                xi_eta_zeta: npt.NDArray[np.float64],
            ) -> npt.NDArray[np.float64]:
                xi, eta, zeta = xi_eta_zeta
                shape_matrix = (
                    1
                    / 8
                    * (1 + xi * corner_signs[:, 0])
                    * (1 + eta * corner_signs[:, 1])
                    * (1 + zeta * corner_signs[:, 2])
                )
                mapped = shape_matrix @ vertices
                return mapped - point

            return inner

        solutions = []
        for point in points:
            point = point
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                initial_guess = (
                    2 * (point - vertices[0]) / (vertices[7] - vertices[0]) - 1
                )
                initial_guess = np.clip(initial_guess, -1, 1)
            np.nan_to_num(initial_guess, copy=False)
            sol = scipy.optimize.least_squares(
                residual(point),
                initial_guess,
                method="trf",
                xtol=tolerance,
                ftol=tolerance,
                gtol=tolerance,
            )
            if not sol.success:
                solutions.append(False)
            else:
                solutions.append(
                    bool(
                        np.all(np.abs(sol.x) <= 1.0 + tolerance)
                        and np.linalg.norm(residual(point)(sol.x)) <= tolerance
                    )
                )
        return np.array(solutions, dtype=np.bool_)

    def _pillars_z_plane_intersection(self, z: np.float32) -> npt.NDArray[np.float32]:
        shape = self.coord.shape
        coord = self.coord.reshape(shape[0] * shape[1], shape[2] * shape[3])
        x1, y1, z1, x2, y2, z2 = coord.T
        t = (z - z1) / (z2 - z1)

        # Compute x and y for all lines
        x = x1 + t * (x2 - x1)
        y = y1 + t * (y2 - y1)

        # Result: (x, y) coordinates for all lines at z
        result = np.column_stack((x, y))
        return result.reshape(shape[0], shape[1], 2)



"""
Some applications use a colon separated list, called a summarykey,
of the required properties needed to uniquely specify a summary vector.

What properties are required is specified in `OPM Flow manual`_ section
F.9.2. Summary variables are described in the `OPM Flow manual`_
section 11.1.

A summary vector is uniquely specified by giving a summary variable, and
potentially one or more of the following properties: well name, region name, lgr
name, block index, completion index, network name.

For example for field variables, no additional information is required so the
summary key is just the variable name: `FOPR`, `FWPR`, etc. For well variables, the
well name need to be specified: `WOPR:WELL_NAME`, `WAQR:MY_WELL` etc. For
block variables, the index has to be given: `BOPR:10,9,50`. For a local completion,
both the lgr name, well name, and index has to be given: `LWWITH:LGR1:WELL2:3,5,5`.

.. _OPM Flow manual: https://opm-project.org/wp-content/uploads/2023/06/OPM_Flow_Reference_Manual_2023-04_Rev-0_Reduced.pdf
"""

from __future__ import annotations

import re
from enum import Enum, auto
from typing import TypeVar, assert_never

SPECIAL_KEYWORDS = [
    "NAIMFRAC",
    "NBAKFL",
    "NBYTOT",
    "NCPRLINS",
    "NEWTFL",
    "NEWTON",
    "NLINEARP",
    "NLINEARS",
    "NLINSMAX",
    "NLINSMIN",
    "NLRESMAX",
    "NLRESSUM",
    "NMESSAGE",
    "NNUMFL",
    "NNUMST",
    "NTS",
    "NTSECL",
    "NTSMCL",
    "NTSPCL",
    "ELAPSED",
    "MAXDPR",
    "MAXDSO",
    "MAXDSG",
    "MAXDSW",
    "STEPTYPE",
    "WNEWTON",
]


class SummaryKeyType(Enum):
    """Summary keys are divided into types based on summary variable name."""

    AQUIFER = auto()
    BLOCK = auto()
    COMPLETION = auto()
    FIELD = auto()
    GROUP = auto()
    LOCAL_BLOCK = auto()
    LOCAL_COMPLETION = auto()
    LOCAL_WELL = auto()
    NETWORK = auto()
    SEGMENT = auto()
    WELL = auto()
    REGION = auto()
    INTER_REGION = auto()
    OTHER = auto()

    @classmethod
    def from_variable(cls, summary_variable: str) -> SummaryKeyType:
        """Returns the type corresponding to the given summary variable

        >>> SummaryKeyType.from_variable("FOPR").name
        'FIELD'
        >>> SummaryKeyType.from_variable("LWWIT").name
        'LOCAL_WELL'
        """
        KEYWORD_TYPE_MAPPING = {
            "A": cls.AQUIFER,
            "B": cls.BLOCK,
            "C": cls.COMPLETION,
            "F": cls.FIELD,
            "G": cls.GROUP,
            "LB": cls.LOCAL_BLOCK,
            "LC": cls.LOCAL_COMPLETION,
            "LW": cls.LOCAL_WELL,
            "N": cls.NETWORK,
            "S": cls.SEGMENT,
            "W": cls.WELL,
        }
        if not summary_variable:
            raise InvalidSummaryKeyError("Got empty summary keyword")
        if any(special in summary_variable for special in SPECIAL_KEYWORDS):
            return cls.OTHER
        if summary_variable[0] in KEYWORD_TYPE_MAPPING:
            return KEYWORD_TYPE_MAPPING[summary_variable[0]]
        if summary_variable[0:2] in KEYWORD_TYPE_MAPPING:
            return KEYWORD_TYPE_MAPPING[summary_variable[0:2]]
        if summary_variable == "RORFR":
            return cls.REGION

        if any(
            re.fullmatch(pattern, summary_variable)
            for pattern in [r"R.FT.*", r"R..FT.*", r"R.FR.*", r"R..FR.*", r"R.F"]
        ):
            return cls.INTER_REGION
        if summary_variable[0] == "R":
            return cls.REGION

        return cls.OTHER


def is_rate(summary_variable: str) -> bool:
    """Whether the given summary variable is a rate.

    See `opm flow reference manual
    <https://opm-project.org/wp-content/uploads/2023/06/OPM_Flow_Reference_Manual_2023-04_Rev-0_Reduced.pdf>`
    table 11.4 for details.
    """
    match SummaryKeyType.from_variable(summary_variable):
        case (
            SummaryKeyType.WELL
            | SummaryKeyType.GROUP
            | SummaryKeyType.FIELD
            | SummaryKeyType.REGION
            | SummaryKeyType.COMPLETION
        ):
            return _match_rate_root(1, _rate_roots, summary_variable)
        case (
            SummaryKeyType.LOCAL_WELL
            | SummaryKeyType.LOCAL_COMPLETION
            | SummaryKeyType.NETWORK
        ):
            return _match_rate_root(2, _rate_roots, summary_variable)
        case SummaryKeyType.SEGMENT:
            return _match_rate_root(1, _segment_rate_roots, summary_variable)
        case SummaryKeyType.INTER_REGION:
            # Region to region rates are identified by R*FR or R**FR
            return _match_rate_root(2, ["FR"], summary_variable) or _match_rate_root(
                3, ["FR"], summary_variable
            )

    return False


def history_key(key: str) -> str:
    """The history summary key responding to given summary key

    >>> history_key("FOPR")
    'FOPRH'
    >>> history_key("BPR:1,3,8")
    'BPRH:1,3,8'
    >>> history_key("LWWIT:WNAME:LGRNAME")
    'LWWITH:WNAME:LGRNAME'
    """

    # Note that this function is not idempotent and only ad-hoc.
    # It is possible to create to make a better version by looking
    # at opm-flow-manual 2023-04 table 11.8, 11.9, 11.14 11.9
    keyword, *rest = key.split(":")
    return ":".join([keyword + "H", *rest])


class InvalidSummaryKeyError(ValueError):
    pass


def make_summary_key(
    keyword: str,
    number: int | None = None,
    name: str | None = None,
    nx: int | None = None,
    ny: int | None = None,
    lgr_name: str | None = None,
    li: int | None = None,
    lj: int | None = None,
    lk: int | None = None,
) -> str:
    """Converts values found in the um to the summary_key format.

    >>> make_summary_key(keyword="WOPR", name="WELL1")
    'WOPR:WELL1'
    >>> make_summary_key(keyword="BOPR", number=4, nx=2, ny=2)
    'BOPR:2,2,1'


    Args:
        keyword: Summary variable name (e.g., ``"WOPR"``, ``"BPR"``).
        number: Numeric qualifier from ``NUMS`` (cell index, region id, etc.).
        name: Text qualifier from ``WGNAMES`` (well/group name).
        nx: Grid dimension in x for block/completion keys.
        ny: Grid dimension in y for block/completion keys.
        lgr_name: Local grid name for local keys.
        li: Local i-index for local block/completion.
        lj: Local j-index for local block/completion.
        lk: Local k-index for local block/completion.

    Raises:
        InvalidSummaryKeyError: If the key is invalid
    """
    match SummaryKeyType.from_variable(keyword):
        case SummaryKeyType.FIELD | SummaryKeyType.OTHER:
            return keyword
        case SummaryKeyType.REGION:
            (number,) = _check_if_missing("region", "nums", number)
            _check_is_positive_number("region", number)
            return f"{keyword}:{number}"
        case SummaryKeyType.AQUIFER:
            (number,) = _check_if_missing("aquifer", "nums", number)
            _check_is_positive_number("aquifer", number)
            return f"{keyword}:{number}"
        case SummaryKeyType.BLOCK:
            nx, ny = _check_if_missing("block", "dimens", nx, ny)
            (number,) = _check_if_missing("block", "nums", number)
            _check_is_positive_number("block", number)
            i, j, k = _cell_index(number - 1, nx, ny)
            return f"{keyword}:{i},{j},{k}"
        case SummaryKeyType.WELL:
            (name,) = _check_if_missing("well", "name", name)
            _check_if_valid_name("well", name)
            return f"{keyword}:{name}"
        case SummaryKeyType.GROUP:
            (name,) = _check_if_missing("group", "name", name)
            _check_if_valid_name("group", name)
            return f"{keyword}:{name}"
        case SummaryKeyType.SEGMENT:
            (name,) = _check_if_missing("segment", "name", name)
            _check_if_valid_name("segment", name)
            (number,) = _check_if_missing("segment", "nums", number)
            _check_is_positive_number("segment", number)
            return f"{keyword}:{name}:{number}"
        case SummaryKeyType.COMPLETION:
            nx, ny = _check_if_missing("completion", "dimens", nx, ny)
            (number,) = _check_if_missing("completion", "nums", number)
            _check_is_positive_number("completion", number)
            (name,) = _check_if_missing("completion", "name", name)
            _check_if_valid_name("completion", name)
            i, j, k = _cell_index(number - 1, nx, ny)
            return f"{keyword}:{name}:{i},{j},{k}"
        case SummaryKeyType.INTER_REGION:
            (number,) = _check_if_missing("inter region", "nums", number)
            _check_is_positive_number("inter region", number)
            r1 = number % 32768
            r2 = ((number - r1) // 32768) - 10
            return f"{keyword}:{r1}-{r2}"
        case SummaryKeyType.LOCAL_WELL:
            (name,) = _check_if_missing("local well", "WGNAMES", name)
            _check_if_valid_name("local well", name)
            (lgr_name,) = _check_if_missing("local well", "LGRS", lgr_name)
            return f"{keyword}:{lgr_name}:{name}"
        case SummaryKeyType.LOCAL_BLOCK:
            li, lj, lk = _check_if_missing("local block", "NUMLX", li, lj, lk)
            (lgr_name,) = _check_if_missing("local block", "LGRS", lgr_name)
            return f"{keyword}:{lgr_name}:{li},{lj},{lk}"
        case SummaryKeyType.LOCAL_COMPLETION:
            (name,) = _check_if_missing("local completion", "WGNAMES", name)
            _check_if_valid_name("local completion", name)
            li, lj, lk = _check_if_missing("local completion", "NUMLX", li, lj, lk)
            (lgr_name,) = _check_if_missing("local completion", "LGRS", lgr_name)
            return f"{keyword}:{lgr_name}:{name}:{li},{lj},{lk}"
        case SummaryKeyType.NETWORK:
            (name,) = _check_if_missing("network", "WGNAMES", name)
            return f"{keyword}:{name}"
        case default:
            assert_never(default)


T = TypeVar("T")

__all__ = [
    "SummaryKeyType",
    "history_key",
    "is_rate",
    "make_summary_key",
    "InvalidSummaryKeyError",
]


def _check_if_missing(
    keyword_name: str, missing_key: str, *test_vars: T | None
) -> list[T]:
    if any(v is None for v in test_vars):
        raise InvalidSummaryKeyError(f"{keyword_name} keyword without {missing_key}")
    return test_vars  # type: ignore


_DUMMY_NAME = ":+:+:+:+"


def _check_if_valid_name(keyword_name: str, name: str) -> None:
    if not name or name == _DUMMY_NAME:
        raise InvalidSummaryKeyError(
            f"{keyword_name} keyword given invalid name '{name}'"
        )


def _check_is_positive_number(keyword_name: str, number: int) -> None:
    if number < 0:
        raise InvalidSummaryKeyError(
            f"{keyword_name} keyword given negative number {number}"
        )


def _cell_index(array_index: int, nx: int, ny: int) -> tuple[int, int, int]:
    """Convert a flat (0-based) index to 1-based (i, j, k) grid indices.

    Args:
        array_index: Zero-based flat index into a grid laid
            out as ``k`` layers of ``ny*nx``.
        nx: Number of cells in the x-direction (strictly positive).
        ny: Number of cells in the y-direction (strictly positive).

    Returns:
        A tuple ``(i, j, k)`` where each component is **1-based**.
    """
    k = array_index // (nx * ny)
    array_index -= k * (nx * ny)
    j = array_index // nx
    array_index -= j * nx
    return array_index + 1, j + 1, k + 1


_rate_roots = [  # see opm-flow-manual 2023-04 table 11.8, 11.9 & 11.14
    "OPR",
    "OIR",
    "OVPR",
    "OVIR",
    "OFR",
    "OPP",
    "OPI",
    "OMR",
    "GPR",
    "GIR",
    "GVPR",
    "GVIR",
    "GFR",
    "GPP",
    "GPI",
    "GMR",
    "WGPR",
    "WGIR",
    "WPR",
    "WIR",
    "WVPR",
    "WVIR",
    "WFR",
    "WPP",
    "WPI",
    "WMR",
    "LPR",
    "LFR",
    "VPR",
    "VIR",
    "VFR",
    "GLIR",
    "RGR",
    "EGR",
    "EXGR",
    "SGR",
    "GSR",
    "FGR",
    "GIMR",
    "GCR",
    "NPR",
    "NIR",
    "CPR",
    "CIR",
    "SIR",
    "SPR",
    "TIR",
    "TPR",
    "GOR",  # dimensionless but considered a rate, as the ratio of two rates
    "WCT",  # dimensionless but considered a rate, as the ratio of two rates
    "OGR",  # dimensionless but considered a rate, as the ratio of two rates
    "WGR",  # dimensionless but considered a rate, as the ratio of two rates
    "GLR",  # dimensionless but considered a rate, as the ratio of two rates
]

_segment_rate_roots = [  # see opm-flow-manual 2023-04 table 11.19
    "OFR",
    "GFR",
    "WFR",
    "CFR",
    "SFR",
    "TFR",
    "CVPR",
    "WCT",  # dimensionless but considered a rate, as the ratio of two rates
    "GOR",  # dimensionless but considered a rate, as the ratio of two rates
    "OGR",  # dimensionless but considered a rate, as the ratio of two rates
    "WGR",  # dimensionless but considered a rate, as the ratio of two rates
]


def _match_rate_root(start: int, rate_roots: list[str], keyword: str) -> bool:
    if len(keyword) < start:
        return False
    return any(keyword[start:].startswith(key) for key in rate_roots)



"""
The egrid fileformat is a file used by reservoir simulators such as opm
flow containing the grid geometry.

For details about the data format see https://resfo.readthedocs.io/en/latest/the_file_format.html


The basic usage is to use the ``egrids`` generator::

    from hypothesis import given
    from resfo_utilities import egrids, EGrid

    @given(egrids)
    def test_egrid(egrid: EGrid):
        print(egrid.shape) # tuple ni,nj,nk
        egrid.to_file("MY_CASE.EGRID")


egrid files contain tuples of keywords and list of data values
of one type (An array with a name). The enums in this file generally describe
a range of values for a position in one of these lists, the dataclasses describe
the values of one keyword or a collection of those, named a file section.

The following egrid file contents (as keyword/array pairs)::

  ("FILEHEAD", [2001,3,0,3,0,0,0])
  ("GRIDUNIT", "METRES   ")

is represented by::

    EGrid(
        Filehead(2001,3,3,TypeOfGrid.CORNER_POINT,RockModel(0),GridFormat(0)),
        GridUnit("METRES   ")
    )

Where ``EGrid`` is a section of the file, ``Filehead`` and ``GridUnit`` are
keywords.

Generally, the data layout of these objects map 1-to-1 with some section of an
valid egrid file.

keywords implement the `to_egrid` that convert from the object representation
to the in file representation.
"""

from dataclasses import astuple, dataclass
from enum import Enum, auto, unique
from typing import Any, assert_never, IO
from os import PathLike

import hypothesis.strategies as st
import numpy as np
import resfo
from hypothesis.extra.numpy import arrays
import numpy.typing as npt


@unique
class Units(Enum):
    """The Grids distance units."""

    METRES = auto()
    CM = auto()
    FEET = auto()

    def to_egrid(self) -> str:
        return self.name.ljust(8)


@unique
class GridRelative(Enum):
    """GridRelative is the second value given GRIDUNIT keyword.

    MAP means map relative units, while
    leaving it blank means relative to the origin given by the
    MAPAXES keyword.
    """

    MAP = auto()
    ORIGIN = auto()

    def to_egrid(self) -> str:
        if self == GridRelative.MAP:
            return "MAP".ljust(8)
        else:
            return "".ljust(8)


@dataclass
class GrdeclKeyword:
    """An abstract grdecl keyword.

    Gives a general implementation of to/from grdecl which recurses on
    fields. Ie. a dataclass such as
    >>> class A(GrdeclKeyword):
    ...     ...
    >>> class B(GrdeclKeyword):
    ...     ...

    >>> @dataclass
    ... class MyKeyword(GrdeclKeyword):
    ...     field1: A
    ...     field2: B

    will have a to_egrid method that will be similar to

    >>> def to_egrid(self):
    ...     return [self.field1.to_egrid(), self.field2.to_egrid]
    """

    def to_egrid(self) -> list[Any]:
        return [value.to_egrid() for value in astuple(self)]


@dataclass
class GridUnit(GrdeclKeyword):
    """Defines the units used for grid dimensions.

    The first value is a string describing the units used, defaults to METRES,
    known accepted other units are FIELD and LAB. The last value describes
    whether the measurements are relative to the map or to the origin of
    MAPAXES.
    """

    unit: Units = Units.METRES
    grid_relative: GridRelative = GridRelative.ORIGIN


@unique
class CoordinateType(Enum):
    """The coordinate system type given in the SPECGRID keyword.

    This is given by either T or F in the last value of SPECGRID, meaning
    either cylindrical or cartesian coordinates respectively.
    """

    CARTESIAN = auto()
    CYLINDRICAL = auto()

    def to_egrid(self) -> int:
        if self == CoordinateType.CARTESIAN:
            return 0
        else:
            return 1


@unique
class TypeOfGrid(Enum):
    """
    A Grid has three possible data layout formats, UNSTRUCTURED, CORNER_POINT,
    BLOCK_CENTER and COMPOSITE (meaning combination of the two former).
    """

    COMPOSITE = 0
    CORNER_POINT = 1
    UNSTRUCTURED = 2
    BLOCK_CENTER = 3

    @property
    def alternate_value(self) -> int:
        """Inverse of alternate_code."""
        alternate_value = 0
        match self:
            case TypeOfGrid.CORNER_POINT:
                alternate_value = 0
            case TypeOfGrid.UNSTRUCTURED:
                alternate_value = 1
            case TypeOfGrid.COMPOSITE:
                alternate_value = 2
            case TypeOfGrid.BLOCK_CENTER:
                alternate_value = 3

            case default:
                assert_never(default)
        return alternate_value


@unique
class RockModel(Enum):
    """Type of rock model."""

    SINGLE_PERMEABILITY_POROSITY = 0
    DUAL_POROSITY = 1
    DUAL_PERMEABILITY = 2


@unique
class GridFormat(Enum):
    """
    The format of the "original grid", ie., what
    method was used to construct the values in the file.
    """

    UNKNOWN = 0
    IRREGULAR_CORNER_POINT = 1
    REGULAR_CARTESIAN = 2


@dataclass
class Filehead:
    """
    The first keyword in an egrid file is the FILEHEAD
    keyword, containing metadata about the file and its
    content.
    """

    version_number: int
    year: int
    version_bound: int
    type_of_grid: TypeOfGrid
    rock_model: RockModel
    grid_format: GridFormat

    def to_egrid(self) -> np.ndarray:
        """
        Returns:
            List of values, as layed out after the FILEHEAD keyword for
            the given filehead.
        """
        # The data is expected to consist of
        # 100 integers, but only a subset is used.
        result = np.zeros((100,), dtype=np.int32)
        result[0] = self.version_number
        result[1] = self.year
        result[3] = self.version_bound
        result[4] = self.type_of_grid.alternate_value
        result[5] = self.rock_model.value
        result[6] = self.grid_format.value
        return result


@dataclass
class GridHead:
    """
    Both for lgr (see LGRSection) and the global grid (see GlobalGrid)
    the GRIDHEAD keyword indicates the start of the grid layout for that
    section.
    """

    type_of_grid: TypeOfGrid
    num_x: int
    num_y: int
    num_z: int
    grid_reference_number: int
    numres: int
    nseg: int
    coordinate_type: CoordinateType
    lgr_start: tuple[int, int, int]
    lgr_end: tuple[int, int, int]

    def to_egrid(self) -> np.ndarray:
        # The data is expected to consist of
        # 100 integers, but only a subset is used.
        result = np.zeros((100,), dtype=np.int32)
        result[0] = self.type_of_grid.value
        result[1] = self.num_x
        result[2] = self.num_y
        result[3] = self.num_z
        result[4] = self.grid_reference_number
        result[24] = self.numres
        result[25] = self.nseg
        result[26] = self.coordinate_type.to_egrid()
        result[[27, 28, 29]] = np.array(self.lgr_start)
        result[[30, 31, 32]] = np.array(self.lgr_end)
        return result


@dataclass
class GlobalGrid:
    """
    The global grid contains the layout of the grid before
    refinements, and the sectioning into grid coarsening
    through the optional corsnum keyword.
    """

    grid_head: GridHead
    coord: np.ndarray
    zcorn: np.ndarray
    actnum: np.ndarray | None = None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GlobalGrid):
            return False
        if self.actnum is None:
            return other.actnum is None
        if other.actnum is None:
            return self.actnum is None
        return (
            self.grid_head == other.grid_head
            and np.array_equal(self.actnum, other.actnum)
            and np.array_equal(self.coord, other.coord)
            and np.array_equal(self.zcorn, other.zcorn)
        )

    def to_egrid(self) -> list[tuple[str, Any]]:
        result = [
            ("GRIDHEAD", self.grid_head.to_egrid()),
            ("COORD   ", self.coord.astype(np.float32)),
            ("ZCORN   ", self.zcorn.astype(np.float32)),
        ]
        if self.actnum is not None:
            result.append(("ACTNUM  ", self.actnum.astype(np.int32)))
        result.append(("ENDGRID ", np.array([], dtype=np.int32)))
        return result


@dataclass
class EGrid:
    """Contains the data of an EGRID file.

    Args:
        file_head:
            The file header starting with the FILEHEAD keyword
        global_grid:
            The global grid
    """

    file_head: Filehead
    grid_unit: GridUnit
    global_grid: GlobalGrid

    @property
    def shape(self) -> tuple[int, int, int]:
        grid_head = self.global_grid.grid_head
        return (grid_head.num_x, grid_head.num_y, grid_head.num_z)

    def to_file(self, filelike: str | PathLike[str] | IO[Any]) -> None:
        """write the EGrid to file.

        Args:
            filelike (str,Path,stream): The egrid file to write to.
        """
        contents = []
        contents.append(("FILEHEAD", self.file_head.to_egrid()))
        contents.append(("GRIDUNIT", self.grid_unit.to_egrid()))  # type: ignore
        contents += self.global_grid.to_egrid()
        resfo.write(filelike, contents)


_finites = st.floats(
    min_value=-100.0, max_value=100.0, allow_nan=False, allow_infinity=False, width=32
)

_indices = st.integers(min_value=1, max_value=4)


def _zcorns(dims: tuple[int, int, int]) -> st.SearchStrategy[npt.NDArray[Any]]:
    return arrays(
        shape=8 * dims[0] * dims[1] * dims[2],
        dtype=np.float32,
        elements=_finites,
    )


_types_of_grid = st.just(TypeOfGrid.CORNER_POINT)
_file_heads = st.builds(
    Filehead,
    st.integers(min_value=0, max_value=5),
    st.integers(min_value=2000, max_value=2022),
    st.integers(min_value=0, max_value=5),
    _types_of_grid,
    grid_format=st.just(GridFormat.IRREGULAR_CORNER_POINT),
)

_grid_heads = st.builds(
    GridHead,
    _types_of_grid,
    _indices,
    _indices,
    _indices,
    _indices,
    st.just(1),
    st.just(1),
    lgr_start=st.tuples(_indices, _indices, _indices),
    lgr_end=st.tuples(_indices, _indices, _indices),
)


@st.composite
def _global_grids(draw: st.DrawFn) -> GlobalGrid:
    grid_head = draw(_grid_heads)
    dims = (grid_head.num_x, grid_head.num_y, grid_head.num_z)
    corner_size = (dims[0] + 1) * (dims[1] + 1) * 6
    coord = arrays(
        shape=corner_size,
        dtype=np.float32,
        elements=_finites,
    )
    actnum = st.one_of(
        st.just(None),
        arrays(
            shape=dims[0] * dims[1] * dims[2],
            dtype=np.int32,
            elements=st.integers(min_value=0, max_value=3),
        ),
    )
    return GlobalGrid(
        coord=draw(coord),
        zcorn=draw(_zcorns(dims)),
        actnum=draw(actnum),
        grid_head=grid_head,
    )


egrids = st.builds(EGrid, _file_heads, global_grid=_global_grids())

__all__ = [
    "GrdeclKeyword",
    "Units",
    "GridRelative",
    "GridUnit",
    "CoordinateType",
    "TypeOfGrid",
    "RockModel",
    "GridFormat",
    "Filehead",
    "GridHead",
    "GlobalGrid",
    "EGrid",
    "egrids",
]



"""
The testing module implements hypothesis generators for data commonly
found in reservoir simulator output.

The basic usage is to use either the ``egrids`` generator::

    from hypothesis import given
    from resfo_utilities import egrids, EGrid

    @given(egrids)
    def test_egrid(egrid: EGrid):
        print(egrid.shape) # tuple ni,nj,nk
        egrid.to_file("MY_CASE.EGRID")

or the ``summaries`` generator::

    from resfo_utilities import SummaryReader
    from resfo_utilities.testing import summaries
    from io import BytesIO
    from hypothesis import given

    @given(summary=summaries())
    def test_that_the_read_values_matches_those_in_the_input(summary):
        smspec, unsmry = summary
        smspec_buf = BytesIO()
        unsmry_buf = BytesIO()
        smspec.to_file(smspec_buf)
        unsmry.to_file(unsmry_buf)
        smspec_buf.seek(0)
        unsmry_buf.seek(0)

        summary = SummaryReader(smspec=lambda: smspec_buf, summaries=[lambda: unsmry_buf])
"""

from ._egrid_generator import (
    Units,
    GridRelative,
    GridUnit,
    CoordinateType,
    TypeOfGrid,
    RockModel,
    GridFormat,
    Filehead,
    GridHead,
    GlobalGrid,
    EGrid,
    egrids,
)

from ._summary_generator import (
    summary_variables,
    UnitSystem,
    Simulator,
    SmspecIntehead,
    Date,
    Smspec,
    smspecs,
    SummaryMiniStep,
    SummaryStep,
    Unsmry,
    summaries,
)

__all__ = [
    "Units",
    "GridRelative",
    "GridUnit",
    "CoordinateType",
    "TypeOfGrid",
    "RockModel",
    "GridFormat",
    "Filehead",
    "GridHead",
    "GlobalGrid",
    "EGrid",
    "egrids",
    "summary_variables",
    "UnitSystem",
    "Simulator",
    "SmspecIntehead",
    "Date",
    "Smspec",
    "smspecs",
    "SummaryMiniStep",
    "SummaryStep",
    "Unsmry",
    "summaries",
]



from resfo_utilities import CornerpointGrid
import numpy as np
from itertools import product
import pytest


@pytest.fixture
def large_regular_grid():
    ni, nj, nk = 50, 50, 10
    height = 100
    top_depth = 0.0
    bot_depth = top_depth + height
    coord = np.zeros((ni + 1, nj + 1, 2, 3), dtype=np.float32)
    zcorn = np.zeros((ni, nj, nk, 8), dtype=np.float32)
    for i, j in product(range(ni + 1), range(nj + 1)):
        coord[i, j, 0] = [i, j, top_depth]
        coord[i, j, 1] = [i, j, bot_depth]
    for i, j, k in product(range(ni), range(nj), range(nk)):
        zcorn[i, j, k] = [height * (k / nk)] * 4 + [height * ((k + 1) / nk)] * 4
    return CornerpointGrid(coord, zcorn)


def test_benchmark_find_cell(large_regular_grid, benchmark):
    def run():
        assert large_regular_grid.find_cell_containing_point(
            [(i + 25.5, j + 25.5, 20.5) for i, j in product(range(10), range(10))]
        ) == [(i + 25, j + 25, 2) for i, j in product(range(10), range(10))]

    benchmark(run)



from resfo_utilities import (
    CornerpointGrid,
    InvalidEgridFileError,
    MapAxes,
    InvalidGridError,
)
from resfo_utilities.testing import egrids
import resfo
import pytest
from io import BytesIO
import numpy as np
from numpy.testing import assert_allclose
from hypothesis import given, assume, example
import hypothesis.strategies as st
from hypothesis.extra.numpy import arrays, from_dtype
from itertools import product
from contextlib import suppress


def write_to_buffer(file_contents):
    buffer = BytesIO()
    resfo.write(buffer, file_contents)
    buffer.seek(0)
    return buffer


def pad_to(lst: list[int], target_len: int):
    return np.pad(lst, (0, target_len - len(lst)), mode="constant")


@given(egrids)
def test_that_cornerpoint_grid_reads_all_generated_grids(egrid):
    buf = BytesIO()
    egrid.to_file(buf)
    buf.seek(0)
    with suppress(InvalidEgridFileError):  # for radial grids
        _ = CornerpointGrid.read_egrid(buf)


def test_that_read_egrid_raises_invalid_egrid_file_when_gridhead_is_mess():
    with pytest.raises(InvalidEgridFileError, match="MESS"):
        CornerpointGrid.read_egrid(write_to_buffer([("GRIDHEAD", resfo.MESS)]))


def test_that_read_egrid_raises_invalid_egrid_file_when_gridhead_is_too_short():
    with pytest.raises(InvalidEgridFileError, match="contained too few elements"):
        CornerpointGrid.read_egrid(write_to_buffer([("GRIDHEAD", [1.0])]))


def test_that_read_egrid_raises_invalid_egrid_file_when_coord_is_mess():
    with pytest.raises(InvalidEgridFileError, match="MESS"):
        CornerpointGrid.read_egrid(
            write_to_buffer([("GRIDHEAD", [1, 1, 1, 1]), ("COORD   ", resfo.MESS)])
        )


def test_that_read_egrid_raises_invalid_egrid_file_the_coordinate_system_is_radial():
    grid_head_array = pad_to([1, 1, 1, 1], 100)
    grid_head_array[26] = 1  # set to 1 to indicate radial grid
    with pytest.raises(InvalidEgridFileError, match="contains a radial grid"):
        CornerpointGrid.read_egrid(write_to_buffer([("GRIDHEAD", grid_head_array)]))


def test_that_read_egrid_warns_when_the_global_grid_does_not_have_reference_number_zero():
    with pytest.warns(UserWarning, match="reference number 1, expected 0"):
        CornerpointGrid.read_egrid(
            write_to_buffer(
                [
                    ("GRIDHEAD", [1, 1, 1, 1, 1]),
                    ("COORD   ", [1.0] * (8 * 3)),
                    ("ZCORN   ", [1.0] * 8),
                ],
            )
        )


def test_that_read_egrid_raises_invalid_egrid_file_when_coord_has_too_many_values():
    with pytest.raises(InvalidEgridFileError, match="did not match grid dimensions"):
        CornerpointGrid.read_egrid(
            write_to_buffer(
                [
                    ("GRIDHEAD", [1, 1, 1, 1]),
                    ("COORD   ", [1.0]),
                    ("ZCORN   ", [1.0] * 8),
                ],
            )
        )


def test_that_read_egrid_raises_invalid_egrid_file_when_mapaxes_is_mess():
    with pytest.raises(InvalidEgridFileError, match="MESS"):
        CornerpointGrid.read_egrid(write_to_buffer([("MAPAXES ", resfo.MESS)]))


def test_that_read_egrid_raises_invalid_egrid_file_when_mapaxes_has_too_many_values():
    with pytest.raises(InvalidEgridFileError, match="contained too few elements"):
        CornerpointGrid.read_egrid(write_to_buffer([("MAPAXES ", [1.0])]))


def test_that_grid_with_invalid_coord_shape_raises():
    with pytest.raises(InvalidGridError, match="coord had invalid dimensions"):
        CornerpointGrid(
            coord=np.array([], dtype=np.float32),
            zcorn=np.array([[[[0, 0, 0, 0, 1, 1, 1, 1]]]], dtype=np.float32),
        )


def test_that_grid_with_invalid_zcorn_shape_raises():
    with pytest.raises(InvalidGridError, match="zcorn had invalid dimensions"):
        CornerpointGrid(
            coord=np.array(
                [
                    [[[0, 0, 0], [0, 0, 1]], [[0, 1, 0], [0, 1, 1]]],
                    [[[1, 0, 0], [1, 0, 1]], [[1, 1, 0], [1, 1, 1]]],
                ],
                dtype=np.float32,
            ),
            zcorn=np.array([], dtype=np.float32),
        )


def test_that_grid_with_zcorn_and_coord_shape_mismatch_raises():
    with pytest.raises(
        InvalidGridError, match="zcorn and coord dimensions do not match"
    ):
        CornerpointGrid(
            coord=np.array(
                [
                    [[[0, 0, 0], [0, 0, 1]], [[0, 1, 0], [0, 1, 1]]],
                    [[[1, 0, 0], [1, 0, 1]], [[1, 1, 0], [1, 1, 1]]],
                ],
                dtype=np.float32,
            ),
            zcorn=np.array(
                [[[[0, 0, 0, 0, 1, 1, 1, 1]]], [[[0, 0, 0, 0, 1, 1, 1, 1]]]],
                dtype=np.float32,
            ),
        )


@pytest.mark.parametrize(
    "contents_after_global_grid",
    [
        [],
        [
            ("LGR     ", np.array([b"LGR1    "], dtype="|S8")),
            ("LGRPARNT", np.array([b"        "], dtype="|S8")),
            (
                "GRIDHEAD",
                pad_to([1, 2, 2, 2, 1] + [0] * 19 + [1, 1, 0, 2, 2, 2, 2, 2, 2], 100),
            ),
            ("COORD   ", np.arange(200, 254)),
            ("ZCORN   ", np.arange(300, 364)),
            ("ACTNUM  ", np.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=">i4")),
            ("HOSTNUM ", np.array([8, 8, 8, 8, 8, 8, 8, 8], dtype=">i4")),
            ("ENDGRID ", np.array([], dtype=">i4")),
            ("ENDLGR  ", np.array([], dtype=">i4")),
            ("NNCHEAD ", pad_to([0, 1], 10)),
            ("NNC1    ", np.array([], dtype=">i4")),
            ("NNC2    ", np.array([], dtype=">i4")),
            ("NNCL    ", np.array([], dtype=">i4")),
            ("NNCG    ", np.array([], dtype=">i4")),
        ],
    ],
)
def test_that_read_egrid_fetches_the_geometry_from_the_global_grid_in_the_file(
    contents_after_global_grid,
):
    grid = CornerpointGrid.read_egrid(
        write_to_buffer(
            [
                ("FILEHEAD", pad_to([3, 2007, 0, 0, 0, 0, 1], 100)),
                ("MAPAXES ", np.array([0.0, 1.0, 0.0, 0.0, 1.0, 0.0], dtype=">f4")),
                ("GRIDUNIT", np.array([b"METRES  ", b"        "], dtype="|S8")),
                ("GRIDHEAD", pad_to([1, 2, 2, 2], 100)),
                ("COORD   ", np.arange(0, 54, dtype=">f4")),
                ("ZCORN   ", np.arange(100, 164, dtype=">f4")),
                ("ACTNUM  ", np.ones((8,), dtype=">i4")),
                ("ENDGRID ", np.array([], dtype=">i4")),
                *contents_after_global_grid,
            ]
        )
    )
    assert grid.map_axes.origin == (0.0, 0.0)
    assert grid.map_axes.y_axis == (0.0, 1.0)
    assert grid.map_axes.x_axis == (1.0, 0.0)
    # coord indecies are (ni, nj, nk, 3)
    # where each triplet is the x,y,z coordinate
    # of the pillar at (i,j,k)

    # The opm manual describes the order in the input as follows:
    # "COORD defines a set of coordinate lines or pillars for a reservoir grid via
    # an array. A total of 6 x (NX+1) x (NY+1) lines must be specified for each
    # coordinate data set (or reservoir). For multiple reservoirs, where
    # NUMRES is greater than one, there must be 6 x (NX+1) x (NY+1) x NUMRES values.
    # For Cartesian geometry, each line is defined by the (x, y, z) coordinates of
    # two distinct points on the line. The lines are entered with I cycling fastest then J."
    assert grid.coord.tolist() == [
        [
            [[0.0, 1.0, 2.0], [3.0, 4.0, 5.0]],  # Top and base points for i,j=0,0
            [[18.0, 19.0, 20.0], [21.0, 22.0, 23.0]],
            [[36.0, 37.0, 38.0], [39.0, 40.0, 41.0]],
        ],
        [
            [[6.0, 7.0, 8.0], [9.0, 10.0, 11.0]],
            [[24.0, 25.0, 26.0], [27.0, 28.0, 29.0]],
            [[42.0, 43.0, 44.0], [45.0, 46.0, 47.0]],
        ],
        [
            [[12.0, 13.0, 14.0], [15.0, 16.0, 17.0]],
            [[30.0, 31.0, 32.0], [33.0, 34.0, 35.0]],
            [[48.0, 49.0, 50.0], [51.0, 52.0, 53.0]],
        ],
    ]
    # coord indecies are (ni, nj, nk, 8)
    # where the 8 values describe depth of the corners
    # for cell (i,j,k)
    # Order of heights for each corner is
    # (N(orth) means higher y, E(east) means higer x, T(op) means lower z (depth))
    # TSW TSE TNW TNE BSW BSE BNW BNE

    # The opm manual describes the order of the input as follows:
    # "ZCORN defines the depth of each corner point of a grid block
    # on the pillars defining the reservoir grid. A total of 8 x NX x NY x NZ values
    # are needed to fully define all the depths in the model. The depths
    # specifying the top of the first layer are entered first with one point
    # for each pillar for each grid block. The points are entered
    # with the X axis cycling fastest. Next come the depths of the bottom
    # of the first layer. The top of layer two follows etc."
    assert grid.zcorn.tolist() == [
        [
            [
                [100.0, 101.0, 104.0, 105.0, 116.0, 117.0, 120.0, 121.0],
                [132.0, 133.0, 136.0, 137.0, 148.0, 149.0, 152.0, 153.0],
            ],
            [
                [108.0, 109.0, 112.0, 113.0, 124.0, 125.0, 128.0, 129.0],
                [140.0, 141.0, 144.0, 145.0, 156.0, 157.0, 160.0, 161.0],
            ],
        ],
        [
            [
                [102.0, 103.0, 106.0, 107.0, 118.0, 119.0, 122.0, 123.0],
                [134.0, 135.0, 138.0, 139.0, 150.0, 151.0, 154.0, 155.0],
            ],
            [
                [110.0, 111.0, 114.0, 115.0, 126.0, 127.0, 130.0, 131.0],
                [142.0, 143.0, 146.0, 147.0, 158.0, 159.0, 162.0, 163.0],
            ],
        ],
    ]


@pytest.fixture
def unit_cell_grid():
    """A Corner point grid which just contains the unit cube as a cell"""
    return CornerpointGrid(
        coord=np.array(
            [
                [[[0, 0, 0], [0, 0, 1]], [[0, 1, 0], [0, 1, 1]]],
                [[[1, 0, 0], [1, 0, 1]], [[1, 1, 0], [1, 1, 1]]],
            ],
            dtype=np.float32,
        ),
        zcorn=np.array([[[[0, 0, 0, 0, 1, 1, 1, 1]]]], dtype=np.float32),
    )


def test_that_interior_points_are_in_the_cell(unit_cell_grid):
    assert unit_cell_grid.point_in_cell((0.5, 0.5, 0.5), 0, 0, 0)
    assert unit_cell_grid.point_in_cell(
        [(0.5, 0.5, 0.5), (0.25, 0.25, 0.25)], 0, 0, 0
    ).tolist() == [True, True]


def test_that_points_on_corners_are_in_the_cell(unit_cell_grid):
    assert unit_cell_grid.point_in_cell(
        [
            (0.0, 0.0, 0.0),
            (1.0, 0.0, 0.0),
            (0.0, 1.0, 0.0),
            (1.0, 1.0, 0.0),
            (0.0, 0.0, 1.0),
            (1.0, 0.0, 1.0),
            (0.0, 1.0, 1.0),
            (1.0, 1.0, 1.0),
        ],
        0,
        0,
        0,
    ).all()


def test_that_points_on_faces_are_in_the_cell(unit_cell_grid):
    assert unit_cell_grid.point_in_cell(
        [
            (0.0, 0.5, 0.5),
            (1.0, 0.5, 0.5),
            (0.5, 0.0, 0.5),
            (0.5, 1.0, 0.5),
            (0.5, 0.5, 0.0),
            (0.5, 0.5, 1.0),
        ],
        0,
        0,
        0,
    ).all()


def test_that_transform_points_does_not_scale_by_map_axes():
    assert MapAxes((0.0, 10.0), (0.0, 0.0), (1.0, 0.0)).transform_map_points(
        np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 1.0, 1.0]])
    ).tolist() == [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 1.0, 1.0]]

    assert MapAxes((0.0, 1.0), (0.0, 0.0), (10.0, 0.0)).transform_map_points(
        np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 1.0, 1.0]])
    ).tolist() == [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 1.0, 1.0]]


def test_that_transform_points_translates_by_origin():
    assert_allclose(
        MapAxes((100.0, 51.0), (100.0, 50.0), (101.0, 50.0)).transform_map_points(
            np.array([[101.0, 50.0, 0.0], [100.0, 51.0, 0.0], [101.0, 51.0, 1.0]])
        ),
        [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 1.0, 1.0]],
    )


coordinates = st.floats(
    allow_nan=False, allow_infinity=False, min_value=-1.0e9, max_value=1.0e9, width=32
)


@st.composite
def regular_grids(draw):
    ni, nj, nk = draw(st.tuples(*([st.integers(min_value=1, max_value=10)] * 3)))
    height = draw(
        st.floats(
            min_value=16.0,
            max_value=2**32,
            allow_nan=False,
            allow_infinity=False,
            width=32,
        )
    )
    top_depth = draw(
        st.floats(
            min_value=0.0,
            max_value=1000.0,
            allow_nan=False,
            allow_infinity=False,
            width=32,
        )
    )
    bot_depth = top_depth + height
    coord = np.zeros((ni + 1, nj + 1, 2, 3), dtype=np.float32)
    zcorn = np.zeros((ni, nj, nk, 8), dtype=np.float32)
    for i, j in product(range(ni + 1), range(nj + 1)):
        coord[i, j, 0] = [i, j, top_depth]
        coord[i, j, 1] = [i, j, bot_depth]
    for i, j, k in product(range(ni), range(nj), range(nk)):
        zcorn[i, j, k] = [height * (k / nk) + top_depth] * 4 + [
            height * ((k + 1) / nk) + top_depth
        ] * 4
    return CornerpointGrid(coord, zcorn)


points = st.tuples(coordinates, coordinates, coordinates)


@given(
    grid=regular_grids(),
    point=points,
    data=st.data(),
)
def test_that_found_cell_contains_point(grid, point, data):
    (cell,) = grid.find_cell_containing_point(np.array([point], dtype=np.float32))
    if cell is None:
        # select one point to check, the point should be
        # in none of the cells
        i, j, k = data.draw(
            st.tuples(
                st.integers(min_value=0, max_value=grid.zcorn.shape[0] - 1),
                st.integers(min_value=0, max_value=grid.zcorn.shape[1] - 1),
                st.integers(min_value=0, max_value=grid.zcorn.shape[2] - 1),
            )
        )
        assert not grid.point_in_cell(point, i, j, k, tolerance=1e-14)
    else:
        assert grid.point_in_cell(point, *cell)


@given(
    grid=regular_grids(),
    point=points,
    data=st.data(),
)
def test_that_on_regular_grids_point_in_cell_is_the_same_as_in_bounding_box(
    grid, point, data
):
    cell = data.draw(
        st.tuples(
            st.integers(min_value=0, max_value=grid.zcorn.shape[0] - 1),
            st.integers(min_value=0, max_value=grid.zcorn.shape[1] - 1),
            st.integers(min_value=0, max_value=grid.zcorn.shape[2] - 1),
        )
    )
    cell_corners = grid.cell_corners(*cell)
    tolerance = 1e-6
    min_point = cell_corners.min(axis=0) - tolerance
    max_point = cell_corners.max(axis=0) + tolerance

    # avoid points close to the boundary
    assume(np.all(np.abs(point - min_point) >= 0.01))
    assume(np.all(np.abs(max_point - point) >= 0.01))

    assert grid.point_in_cell(point, *cell, tolerance) == (
        np.all(min_point <= point) and np.all(point <= max_point)
    )


def test_that_map_coordinates_parameter_sets_the_coordinate_system_for_points():
    # unit cell grid with map axes
    # translating origin to 100, 100
    grid = CornerpointGrid(
        coord=np.array(
            [
                [[[0, 0, 0], [0, 0, 1]], [[0, 1, 0], [0, 1, 1]]],
                [[[1, 0, 0], [1, 0, 1]], [[1, 1, 0], [1, 1, 1]]],
            ],
            dtype=np.float32,
        ),
        zcorn=np.array([[[[0, 0, 0, 0, 1, 1, 1, 1]]]], dtype=np.float32),
        map_axes=MapAxes((100.0, 101.0), (100.0, 100.0), (101.0, 100.0)),
    )

    # A map contained in cell 0,0,0
    point = np.array([[100.5, 100.5, 0.5]], dtype=np.float32)
    # By default points are in the map coordinate system
    assert grid.find_cell_containing_point(point) == [(0, 0, 0)]

    assert grid.find_cell_containing_point(point, map_coordinates=False) == [None]


def test_that_point_in_cell_correctly_orders_zcorn_against_coord():
    """This is a regression test for a bug where point_in_cell
    did not order coord points correctly against zcorn"""
    grid = CornerpointGrid(
        coord=np.array(
            [
                [
                    [[581, 800, 675], [581, 800, 86]],
                    [[581, 700, 683], [581, 700, 110]],
                ],
                [
                    [[681, 800, 657], [681, 800, 48]],
                    [[681, 700, 670], [681, 700, 68]],
                ],
            ],
            dtype=np.float32,
        ),
        zcorn=np.array(
            [
                [
                    [
                        [729, 707, 733, 718, 735, 712, 738, 723],
                    ]
                ]
            ],
            dtype=np.float32,
        ),
    )

    # A point contained in cell 0,0,0
    point = np.array([[662, 752, 718]], dtype=np.float64)
    # By default points are in the map coordinate system
    assert grid.point_in_cell(point, 0, 0, 0)
    assert grid.find_cell_containing_point(point) == [(0, 0, 0)]


def test_that_point_in_cell_correctly_assign_vertices_to_faces():
    """This is a regression test for a bug where point_in_cell
    assigned vertices to faces in a way that created a hole in the cell surface"""
    grid = CornerpointGrid(
        coord=np.array(
            [
                [
                    [[0, 0, 0], [0, 0, 1]],
                    [[1, 0, 0], [1, 0, 1]],
                ],
                [
                    [[0, 2, 0], [0, 1, 1]],
                    [[1, 1, 0], [1, 2, 1]],
                ],
            ],
            dtype=np.float32,
        ),
        zcorn=np.array(
            [
                [
                    [
                        [0, 0, 0, 0, 1, 1, 1, 1],
                    ]
                ]
            ],
            dtype=np.float32,
        ),
    )

    # A point contained in cell 0,0,0
    point = np.array([[0.9, 1.0, 0.75]], dtype=np.float64)

    assert grid.point_in_cell(point, 0, 0, 0)
    assert grid.find_cell_containing_point(point) == [(0, 0, 0)]


@given(*([st.floats(min_value=-1.0, max_value=2.0)] * 3))
@pytest.mark.parametrize(
    "bottom_heights",
    [
        pytest.param(
            [0.05, 1.0, 1.0, 0.05],
            id="the bottom face has two elevated diagonally opposed corners",
        ),
        pytest.param(
            [0.0, 1.0, 1.0, 1.0],
            id="one bottom corner is collapsed onto the corresponding top corner",
        ),
        pytest.param(
            [0.0, 0.0, 1.0, 1.0],
            id="one bottom edge is collapsed onto the corresponding top edge",
        ),
    ],
)
def test_point_in_cell_considers_cells_as_trilinear_shapes(bottom_heights, x, y, z):
    # All grids are unit grids except for the bottom face
    grid = CornerpointGrid(
        coord=np.array(
            [
                [[[0, 0, 0], [0, 0, 1]], [[0, 1, 0], [0, 1, 1]]],
                [[[1, 0, 0], [1, 0, 1]], [[1, 1, 0], [1, 1, 1]]],
            ],
            dtype=np.float32,
        ),
        zcorn=np.array([[[[0, 0, 0, 0, *bottom_heights]]]], dtype=np.float32),
    )

    def bottom_face_depth(x, y):
        """The depth of the bottom face at x,y by bilinear interpolation"""
        xy_1 = (1 - x) * grid.zcorn[0, 0, 0, 4] + x * grid.zcorn[0, 0, 0, 5]
        xy_2 = (1 - x) * grid.zcorn[0, 0, 0, 6] + x * grid.zcorn[0, 0, 0, 7]
        return (1 - y) * xy_1 + y * xy_2

    tolerance = 1e-6

    def in_bounding_box(point):
        return all(-1 * tolerance <= c <= 1 + tolerance for c in point)

    # avoid points that is very close to the boundary to avoid
    # numerical issues
    assume(np.abs(z) >= 0.01)
    assume(np.abs(bottom_face_depth(x, y) - z) >= 0.01)

    # only the bottom face is different from the bounding box
    # so containment is the same as the conjunction of the two conditions
    assert grid.point_in_cell([x, y, z], 0, 0, 0, tolerance) == (
        z <= bottom_face_depth(x, y) and in_bounding_box((x, y, z))
    )


def test_that_zero_height_pillar_is_invalid():
    grid = CornerpointGrid(
        coord=np.array(
            [
                [
                    [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                    [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                ],
                [
                    [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                    [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                ],
            ],
            dtype=np.float32,
        ),
        zcorn=np.array(
            [[[[0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0]]]], dtype=np.float32
        ),
        map_axes=None,
    )
    with pytest.raises(InvalidGridError, match="Grid contains zero height pillars"):
        grid.cell_corners(0, 0, 0)


def test_that_cells_with_infinite_pillars_are_invalid():
    grid = CornerpointGrid(
        coord=np.array(
            [
                [
                    [[np.inf, np.inf, np.inf], [np.inf, np.inf, np.inf]],
                    [[np.inf, np.inf, np.inf], [np.inf, np.inf, np.inf]],
                ],
                [
                    [[np.inf, np.inf, np.inf], [np.inf, np.inf, np.inf]],
                    [[np.inf, np.inf, np.inf], [np.inf, np.inf, np.inf]],
                ],
            ],
            dtype=np.float32,
        ),
        zcorn=np.array(
            [[[[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]]], dtype=np.float32
        ),
        map_axes=None,
    )

    with pytest.raises(InvalidGridError, match="The corners of the cell"):
        grid.cell_corners(0, 0, 0)


@st.composite
def single_cell_grids(draw):
    nice_elements = dict(
        allow_nan=False, allow_infinity=False, max_value=2**12, min_value=-(2**12)
    )
    coord = draw(
        arrays(
            np.float32,
            (2, 2, 2, 3),
            elements=nice_elements,
        )
    )

    for i in range(coord.shape[0]):
        for j in range(coord.shape[1]):
            bot_pillar_z = np.float32(coord[i, j, 0, 2] + 0.1)
            coord[i, j, 1] = draw(
                from_dtype(
                    np.dtype(np.float32),
                    **{
                        **nice_elements,
                        **dict(min_value=bot_pillar_z, max_value=2**14),
                    },
                )
            )

    grid = CornerpointGrid(
        coord=coord,
        zcorn=draw(arrays(np.float32, (1, 1, 1, 8), elements=nice_elements)),
    )
    try:
        grid.cell_corners(0, 0, 0)
    except InvalidGridError:
        assume(False)
    return grid


@given(single_cell_grids(), points)
@example(
    grid=CornerpointGrid(
        coord=np.array(
            [
                [
                    [[1.0, 1.0, 1.0], [2.0, 2.0, 2.0]],
                    [[1.0, 1.0, 1.0], [2.0, 2.0, 2.0]],
                ],
                [
                    [[1.0, 1.0, 1.0], [2.0, 2.0, 2.0]],
                    [[1.0, 1.0, 1.0], [2.0, 2.0, 2.0]],
                ],
            ],
            dtype=np.float32,
        ),
        zcorn=np.array(
            [[[[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]]], dtype=np.float32
        ),
        map_axes=None,
    ),
    point=(0.0, 0.0, 0.0),
).via("discovered failure")
@example(
    grid=CornerpointGrid(
        coord=np.array(
            [
                [
                    [[0.0, 0.0, -1.0], [0.0, 0.0, 1.0]],
                    [[0.0, 1.0, 0.0], [2.0, 1.0, 2.0]],
                ],
                [
                    [[1.0, 0.0, -1.0], [1.0, 0.0, 1.0]],
                    [[0.0, 1.0, -1.0], [0.0, 1.0, 1.0]],
                ],
            ],
            dtype=np.float32,
        ),
        zcorn=np.array(
            [[[[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]]], dtype=np.float32
        ),
        map_axes=None,
    ),
    point=(0.0, 0.0, 0.0),
).via("constructed grid with bulging face")
@example(
    grid=CornerpointGrid(
        coord=np.array(
            [
                [
                    [[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]],
                    [[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]],
                ],
                [
                    [[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]],
                    [[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]],
                ],
            ],
            dtype=np.float32,
        ),
        zcorn=np.array(
            [[[[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]]], dtype=np.float32
        ),
        map_axes=None,
    ),
    point=(0.0, 0.0, 9.999999747378752e-06),
).via("discovered failure")
def test_that_in_single_cell_grids_found_and_contains_are_the_same(
    grid: CornerpointGrid, point: tuple[float, float, float]
):
    assert bool(grid.find_cell_containing_point([point])[0]) == grid.point_in_cell(
        point, 0, 0, 0
    )



project = "resfo-utilities"
copyright = "2022, Equinor"
author = "Equinor"
release = "1.0.0"


extensions = ["sphinx.ext.autodoc", "sphinx.ext.doctest", "sphinx.ext.intersphinx"]
intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable/", None),
    "python": ("https://docs.python.org/3", None),
}
nitpick_ignore = [
    ("py:class", "np.float32"),
    ("py:class", "numpy.float32"),
    ("py:class", "collections.abc.Buffer"),
    ("py:class", "npt.ArrayLike"),
    ("py:class", "ArrayLike"),
    ("py:class", "npt.NDArray"),
    ("py:class", "np.bool_"),
    ("py:class", "hypothesis.strategies.SearchStrategy"),
    ("py:class", "annotated_types.Gt"),
    ("py:class", "annotated_types.Interval"),
    ("py:class", "resfo.format.Format"),
]
language = "python"
html_theme = "sphinx_rtd_theme"
autodoc_type_aliases = {"npt.ArrayLike": "ArrayLike"}



from resfo_utilities import SummaryReader, InvalidSummaryError
from resfo_utilities.testing import summaries, Unsmry
from pathlib import Path
from io import StringIO, BytesIO
import pytest
from hypothesis import given
import hypothesis.strategies as st
from contextlib import suppress
import resfo
import numpy as np


def test_that_summary_reader_can_be_initialized_with_either_path_or_io(tmp_path: Path):
    (tmp_path / "CASE.FSMSPEC").touch()
    (tmp_path / "CASE.FUNSMRY").touch()
    _ = SummaryReader(case_path=tmp_path / "CASE")
    _ = SummaryReader(smspec=StringIO, summaries=[StringIO])
    with pytest.raises(ValueError):
        _ = SummaryReader(
            case_path=tmp_path,
            smspec=StringIO,
            summaries=[StringIO],
        )
    with pytest.raises(ValueError):
        _ = SummaryReader()


@given(st.binary(), st.binary())
def test_that_summary_reader_only_raises_invalid_summary_error(
    spec: bytes, unsmry: bytes
):
    with suppress(InvalidSummaryError):
        reader = SummaryReader(
            smspec=lambda: BytesIO(spec), summaries=[lambda: BytesIO(unsmry)]
        )
        _ = list(reader.values())


def report_step_value(unsmry: Unsmry, report_step: int, kw_index: int):
    return unsmry.steps[report_step].ministeps[-1].params[kw_index]


def step_value(unsmry: Unsmry, index: int, kw_index: int):
    while index >= 0:
        for step in unsmry.steps:
            if index < len(step.ministeps):
                return step.ministeps[index].params[kw_index]
            index -= len(step.ministeps)


@given(summary=summaries(), report_step_only=st.booleans())
def test_that_the_read_values_matches_those_in_the_input(summary, report_step_only):
    smspec, unsmry = summary
    smspec_buf = BytesIO()
    unsmry_buf = BytesIO()
    smspec.to_file(smspec_buf)
    unsmry.to_file(unsmry_buf)
    smspec_buf.seek(0)
    unsmry_buf.seek(0)

    summary = SummaryReader(smspec=lambda: smspec_buf, summaries=[lambda: unsmry_buf])

    values = list(summary.values(report_step_only))
    getter = report_step_value if report_step_only else step_value
    for kw_index, _ in enumerate(summary.summary_keywords):
        for report_step, val in enumerate(values):
            assert getter(unsmry, report_step, kw_index) == pytest.approx(val[kw_index])


def read_summary(smspec, unsmry, report_step_only=True):
    summary = SummaryReader(smspec=lambda: smspec, summaries=[lambda: unsmry])
    _ = list(summary.values(report_step_only))
    _ = list(summary.summary_keywords)


@pytest.mark.parametrize(
    "spec_contents, smry_contents, error_message",
    [
        (b"", b"", "Keyword startdat missing"),
        (b"1", b"1", "Summary files contained invalid contents"),
        (
            b"\x00\x00\x00\x10FOOOOOOO\x00\x00\x00\x01"
            + b"INTE"
            + b"\x00\x00\x00\x10"
            + (4).to_bytes(4, signed=True, byteorder="big")
            + b"\x00" * 4,
            b"",
            "Keyword startdat missing",
        ),
        (
            b"\x00\x00\x00\x10STARTDAT\x00\x00\x00\x01"
            + b"INTE"
            + b"\x00\x00\x00\x10"
            + (4).to_bytes(4, signed=True, byteorder="big")
            + b"\x00" * 4
            + (4).to_bytes(4, signed=True, byteorder="big"),
            b"",
            "contains invalid STARTDAT",
        ),
    ],
)
def test_that_incorrect_summary_files_raises_informative_errors(
    smry_contents, spec_contents, error_message
):
    smry_buf = BytesIO(smry_contents)
    smry_buf.seek(0)
    spec_buf = BytesIO(spec_contents)
    spec_buf.seek(0)

    with pytest.raises(InvalidSummaryError, match=error_message):
        read_summary(spec_buf, smry_buf)


@given(summaries())
def test_truncated_summary_file_raises_invalidresponsefile(summary):
    smspec, unsmry = summary
    smspec_buf = BytesIO()
    unsmry_buf = BytesIO()
    smspec.to_file(smspec_buf)
    unsmry.to_file(unsmry_buf)
    smspec_buf.seek(0)
    unsmry_buf.seek(0)
    unsmry_buf.truncate(10)

    with pytest.raises(InvalidSummaryError, match=""):
        read_summary(smspec_buf, unsmry_buf)


def test_mess_values_in_summary_files_raises_informative_errors():
    smspec_buf = BytesIO()
    resfo.write(smspec_buf, [("KEYWORDS", resfo.MESS)])
    smspec_buf.seek(0)

    with pytest.raises(InvalidSummaryError, match="has incorrect type MESS"):
        read_summary(smspec_buf, BytesIO())


def test_missing_keywords_in_smspec_raises_informative_error():
    smspec_buf = BytesIO()
    resfo.write(
        smspec_buf,
        [
            ("STARTDAT", np.array([31, 12, 2012, 00], dtype=np.int32)),
            ("UNITS   ", ["ANNUAL  "]),
        ],
    )
    smspec_buf.seek(0)

    with pytest.raises(
        InvalidSummaryError,
        match="Keywords missing",
    ):
        read_summary(smspec_buf, BytesIO())



from dataclasses import dataclass
import os
from collections.abc import Iterable, Iterator
from typing import Any, IO, overload, Callable, TypeAlias
from datetime import datetime
from itertools import tee
import re
from functools import partial
import warnings
from natsort import natsorted
import numpy.typing as npt
import resfo
import numpy as np


class InvalidSummaryError(ValueError):
    """Raised when a given summary file is not valid.

    Can be raised either when the file can't be read (eg. a directory)
    or its contents is not valid.
    """


@dataclass
class SummaryKeyword:
    """One member of the KEYWORDS array.

    Attributes:
        summary_variable:
            The variable name, eg WOPR, or FOPT.
        number:
            A number associated with the keyword,
            eg. for block variables it is the index of the block.
        name:
            A name associated with the keyword,
            eg. for well variables it is the name of the well.
        lgr_name:
            If a local variable then the name of the Local Grid
            Refinement.
        li:
            The i index of the host cell for the LGR.
        lj:
            The j index of the host cell for the LGR.
        lk:
            The k index of the host cell for the LGR.
        unit:
            The units for the value of the keyword, eg. for
            FOPR it may be SM3/DAY.
    """

    summary_variable: str
    number: int | None = None
    name: str | None = None
    lgr_name: str | None = None
    li: int | None = None
    lj: int | None = None
    lk: int | None = None
    unit: str | None = None


FileOpener: TypeAlias = Callable[[], IO[Any]]


class SummaryReader:
    """Reader for summary files.

    The result of running a reservoir simulator is a number of time vectors
    which are written to summary files.

    Each file is opened when the corresponding data is requested so asking for
    the properties of SummaryReader may raise `InvalidSummaryError` if the
    corresponding file or its content is invalid.
    """

    @overload
    def __init__(
        self,
        *,
        case_path: str | os.PathLike[str],
        smspec: None = None,
        summaries: None = None,
    ):
        pass

    @overload
    def __init__(
        self,
        *,
        smspec: FileOpener,
        summaries: Iterable[FileOpener],
        case_path: None = None,
    ):
        pass

    def __init__(
        self,
        *,
        case_path: str | os.PathLike[str] | None = None,
        smspec: FileOpener | None = None,
        summaries: Iterable[FileOpener] | None = None,
    ):
        """
        Args:
            case_path:
                The path to one summary file or the basename
                of several summary files. By giving just the base name,
                eg. `path/to/CASE`, SummaryReader will look for summary
                files named eg. `path/to/CASE.SMSPEC`, `path/to/CASE.UNSMRY`
                `path/to/CASE.FSMSPEC`, `path/to/CASE.S0001`, etc. depending
                on whether the summary is formatted and unified.

                The order in which summary files will looked for in is formatted first
                then unified first lexiographicall, ie.: UNSMRY, Snnnn, FUNSMRY, and
                then Annnn.

                By giving an extension, only summary files that match the given
                formatted or unified combination is looked for, ie. if
                case_path="CASE.UNSMRY" then only "CASE.SMSPEC" and "CASE.UNSMRY"
                will be opened.
        Raises:
            FileNotFoundError:
                If the required summary files for the given case_path
                does not exist.
        """
        if case_path is None and (smspec is None or summaries is None):
            raise ValueError(
                "SummaryReader must be initialized with"
                " either case_path or both smspec and summaries."
            )
        if case_path is not None and (smspec is not None or summaries is not None):
            raise ValueError(
                "SummaryReader must be initialized with"
                " either case_path or smspec and summaries, not both."
            )

        if case_path is not None:
            self._smspec, self._summaries = self._get_file_openers(case_path)
        else:
            assert smspec is not None
            assert summaries is not None
            self._smspec = smspec
            self._summaries = summaries

        self._start_date: datetime | None = None
        self._summary_keywords: list[SummaryKeyword] | None = None
        self._dimensions: tuple[int, int, int] | None = None
        self._restart: str | None = None
        self._have_read_smspec = False

    @property
    def smspec_filename(self) -> str:
        """The filename of the summary spec file.

        e.g. "CASE.SMSPEC"

        """
        return self._spec_filename

    @property
    def summary_filenames(self) -> Iterator[str]:
        """The filename of the summary file(s).

        e.g. ["CASE.UNSMRY"] for unified or
        ["CASE.S0001", "CASE.S0002"] for split.

        """
        return iter(self._summary_filenames)

    @property
    def start_date(self) -> datetime:
        """The start date of the simulation."""
        if self._start_date is not None:
            return self._start_date
        self._start_date, self._summary_keywords, self._dimensions, self._restart = (
            _read_spec(self._smspec)
        )
        self._have_read_smspec = True
        assert self._start_date is not None
        return self._start_date

    @property
    def summary_keywords(self) -> list[SummaryKeyword]:
        """The list of keywords in the summary."""
        if self._summary_keywords is not None:
            return self._summary_keywords
        self._start_date, self._summary_keywords, self._dimensions, self._restart = (
            _read_spec(self._smspec)
        )
        self._have_read_smspec = True
        assert self._summary_keywords is not None
        return self._summary_keywords

    @property
    def dimensions(self) -> tuple[int, int, int] | None:
        """The dimensions of the grid used in the simulation."""
        if self._have_read_smspec:
            return self._dimensions
        self._start_date, self._summary_keywords, self._dimensions, self._restart = (
            _read_spec(self._smspec)
        )
        self._have_read_smspec = True
        return self._dimensions

    @property
    def restart(self) -> str | None:
        """The name of the case the simulation was restarted from (if any)."""
        if self._have_read_smspec:
            return self._restart
        self._start_date, self._summary_keywords, self._dimensions, self._restart = (
            _read_spec(self._smspec)
        )
        self._have_read_smspec = True
        return self._restart

    def values(
        self, report_step_only: bool = True
    ) -> Iterator[npt.NDArray[np.float32]]:
        """Iterate over the values for the summary keywords.

        Args:
            report_step_only: If ``True``, yield only at report steps (``DATES``).
        Yields:
            arrays of the keyword values in the order of
            the summary_keywords.
        Raises:
            InvalidSummaryError:
                If the summary files cannot be read from or contains invalid
                contents.
        """

        last_params = None
        try:
            for smry_opener in self._summaries:
                with smry_opener() as smry:
                    summary_name = _stream_name(smry)

                    def read_params() -> Iterator[npt.NDArray[np.float32]]:
                        nonlocal last_params
                        if last_params is not None:
                            vals = _validate_array(
                                "PARAMS", summary_name, last_params.read_array()
                            )
                            last_params = None
                            yield vals

                    for entry in resfo.lazy_read(smry):
                        kw = entry.read_keyword()
                        if last_params and not report_step_only:
                            yield from read_params()
                        if kw == "PARAMS  ":
                            last_params = entry
                        if report_step_only and kw == "SEQHDR  ":
                            yield from read_params()
                    yield from read_params()
        except OSError as err:
            raise InvalidSummaryError(
                f"Could not read from summary file {err.filename}: {err.strerror}"
            ) from err
        except resfo.ResfoParsingError as err:
            raise InvalidSummaryError(
                f"Summary files contained invalid contents: {err}"
            ) from err

    def _get_file_openers(
        self,
        case_path: str | os.PathLike[str],
    ) -> tuple[FileOpener, Iterable[FileOpener]]:
        self.case_path = case_path
        self._summary_filenames, self._spec_filename = _get_summary_filenames(case_path)
        mode = "rt" if self._spec_filename.lower().endswith("fsmspec") else "rb"

        def opener(s: str | os.PathLike[str]) -> FileOpener:
            def inner() -> IO[Any]:
                return open(os.path.abspath(s), mode)

            return inner

        return (
            opener(self._spec_filename),
            [opener(s) for s in self._summary_filenames],
        )


def _read_spec(
    spec_opener: FileOpener,
) -> tuple[datetime, list[SummaryKeyword], tuple[int, int, int] | None, str | None]:
    """Read an SMSPEC file and return start date, keywords, dimensions and restart

    This function performs validation, determines the index of the
    TIME vector and the unit, and read all available keys.

    Args:
        spec: A function that returns a file-like object for the
              SMSPEC (binary or text depending on format).
        key_patterns: Patterns identifying which keys to keep.

    Returns:
        tuple of the start date, list of summary keywords, dimensions,
        and restart case path.

    Raises:
        InvalidSummaryError: On malformed content (e.g., missing UNITS, STARTDAT, etc.)
            or if parsing of the smspec fails.
    """
    start_date = None
    num_keywords = None
    dimensions = None
    wgnames = None
    spec_name = ""
    try:
        with spec_opener() as spec:
            spec_name = _stream_name(spec)

            arrays: dict[str, npt.NDArray[Any] | None] = dict.fromkeys(
                [
                    "NUMS    ",
                    "KEYWORDS",
                    "NUMLX   ",
                    "NUMLY   ",
                    "NUMLZ   ",
                    "LGRS    ",
                    "UNITS   ",
                    "RESTART ",
                ],
                None,
            )
            for entry in resfo.lazy_read(spec):
                # If we have found all values we are looking for
                # we stop reading
                if all(
                    p is not None
                    for p in [start_date, num_keywords, dimensions, *arrays.values()]
                ):
                    break
                kw = entry.read_keyword()
                if kw in arrays:
                    arrays[kw] = _validate_array(kw, spec_name, entry.read_array())
                # "NAMES   " is an alias for "WGNAMES "
                # if kw is one of either, we set wgnames
                if kw in {"WGNAMES ", "NAMES   "}:
                    wgnames = _validate_array(kw, spec_name, entry.read_array())
                if kw == "DIMENS  ":
                    vals = _validate_array(kw, spec_name, entry.read_array())
                    size = len(vals)
                    num_keywords = vals[0] if size > 0 else None
                    dimensions = tuple(vals[1:4]) if size > 3 else None
                if kw == "STARTDAT":
                    vals = _validate_array(kw, spec_name, entry.read_array())
                    size = len(vals)
                    day = vals[0] if size > 0 else 0
                    month = vals[1] if size > 1 else 0
                    year = vals[2] if size > 2 else 0
                    hour = vals[3] if size > 3 else 0
                    minute = vals[4] if size > 4 else 0
                    microsecond = vals[5] if size > 5 else 0
                    try:
                        start_date = datetime(
                            day=day,
                            month=month,
                            year=year,
                            hour=hour,
                            minute=minute,
                            second=microsecond // 10**6,
                            microsecond=microsecond % 10**6,
                        )
                    except Exception as err:
                        raise InvalidSummaryError(
                            f"SMSPEC {spec_name} contains invalid STARTDAT: {err}"
                        ) from err
    except OSError as err:
        raise InvalidSummaryError(
            f"Could not read from summary spec {err.filename}: {err.strerror}"
        ) from err
    except resfo.resfo.ResfoParsingError as err:
        raise InvalidSummaryError(
            f"Summary spec contained invalid contents: {err}"
        ) from err

    keywords = arrays["KEYWORDS"]
    nums = arrays["NUMS    "]
    numlx = arrays["NUMLX   "]
    numly = arrays["NUMLY   "]
    numlz = arrays["NUMLZ   "]
    lgr_names = arrays["LGRS    "]
    units = arrays["UNITS   "]

    if start_date is None:
        raise InvalidSummaryError(f"Keyword startdat missing in {spec_name}")
    if keywords is None:
        raise InvalidSummaryError(f"Keywords missing in {spec_name}")
    if num_keywords is None:
        num_keywords = len(keywords)
        warnings.warn(
            "SMSPEC did not contain num_keywords in DIMENS."
            f" Using length of KEYWORDS: {num_keywords}."
        )
    elif num_keywords > len(keywords):
        warnings.warn(
            f"number of keywords given in DIMENS {num_keywords} is larger than the "
            f"length of KEYWORDS {len(keywords)}, truncating size to match.",
        )
        num_keywords = len(keywords)

    summary_keywords: list[SummaryKeyword] = []

    def optional_get(arr: npt.NDArray[Any] | None, idx: int) -> Any:
        if arr is None:
            return None
        if len(arr) <= idx:
            return None
        return arr[idx]

    def decode_if_byte(key: bytes | str) -> str:
        return key.decode() if isinstance(key, bytes) else key

    @overload
    def key2str(key: bytes | str) -> str:
        pass

    @overload
    def key2str(key: None) -> None:
        pass

    def key2str(key: bytes | str | None) -> str | None:
        if key is None:
            return None
        return decode_if_byte(key).strip()

    for i in range(num_keywords):
        summary_keywords.append(
            SummaryKeyword(
                summary_variable=key2str(keywords[i]),
                number=optional_get(nums, i),
                name=key2str(optional_get(wgnames, i)),
                lgr_name=key2str(optional_get(lgr_names, i)),
                li=optional_get(numlx, i),
                lj=optional_get(numly, i),
                lk=optional_get(numlz, i),
                unit=key2str(optional_get(units, i)),
            )
        )

    restart_arr = arrays["RESTART "]
    restart = None
    if restart_arr is not None:
        restart = "".join(decode_if_byte(s) for s in restart_arr).strip()
        if restart and not os.path.isabs(restart):
            restart = os.path.join(os.path.dirname(spec_name), restart)

    return (
        start_date,
        summary_keywords,
        dimensions,
        restart,
    )


def _validate_array(
    kw: str, filename: str, vals: npt.NDArray[Any] | resfo.MESS
) -> npt.NDArray[Any]:
    if vals is resfo.MESS or isinstance(vals, resfo.MESS):
        raise InvalidSummaryError(f"{kw.strip()} in {filename} has incorrect type MESS")
    return vals


def _has_extension(path: str, ext: str) -> bool:
    """
    >>> _has_extension("ECLBASE.SMSPEC", "smspec")
    True
    >>> _has_extension("BASE.SMSPEC", "smspec")
    True
    >>> _has_extension("BASE.FUNSMRY", "smspec")
    False
    >>> _has_extension("ECLBASE.smspec", "smspec")
    True
    >>> _has_extension("ECLBASE.tar.gz.smspec", "smspec")
    True

    Args:
        path: File name to check.
        ext: Allowed extension regex.

    Returns:
        ``True`` if the file has any of the extensions in ``exts``.
    """
    if "." not in path:
        return False
    splitted = path.split(".")
    return re.fullmatch(ext, splitted[-1].lower()) is not None


def _is_base_with_extension(base: str, path: str, ext: str) -> bool:
    """
    >>> _is_base_with_extension("ECLBASE", "ECLBASE.SMSPEC", "smspec")
    True
    >>> _is_base_with_extension("ECLBASE", "BASE.SMSPEC", "smspec")
    False
    >>> _is_base_with_extension("ECLBASE", "BASE.FUNSMRY", "smspec")
    False
    >>> _is_base_with_extension("ECLBASE", "ECLBASE.smspec", "smspec")
    True
    >>> _is_base_with_extension("ECLBASE.tar.gz", "ECLBASE.tar.gz.smspec", "smspec")
    True

    Args:
        base: Basename without extension.
        path: Candidate path.
        exts: Allowed extension regex pattern.

    Returns:
        ``True`` if ``path`` is ``base`` with one of ``exts``.
    """
    if "." not in path:
        return False
    splitted = path.split(".")
    return (
        ".".join(splitted[0:-1]) == base
        and re.fullmatch(ext, splitted[-1].lower()) is not None
    )


ANY_SUMMARY_EXTENSION = r"unsmry|smspec|funsmry|fsmspec|s\d\d\d\d|a\d\d\d\d"


def _get_summary_filenames(filepath: str | os.PathLike[str]) -> tuple[list[str], str]:
    directory, file_name = os.path.split(filepath)
    if "." in file_name:
        case_name = ".".join(file_name.split(".")[:-1])
    else:
        case_name = file_name
    specified_formatted = _has_extension(file_name, r"funsmry|fsmspec|a\d\d\d\d")
    specified_unformatted = _has_extension(file_name, r"unsmry|smspec|s\d\d\d\d")
    specified_unified = _has_extension(file_name, "funsmry")
    specified_split = _has_extension(file_name, r"x\d\d\d\d|a\d\d\d\d")
    spec_candidates, smry_candidates = tee(
        map(
            lambda x: os.path.join(directory, x),
            filter(
                lambda x: _is_base_with_extension(
                    path=x, base=case_name, ext=ANY_SUMMARY_EXTENSION
                ),
                os.listdir(directory or "."),
            ),
        )
    )

    def filter_extension(ext: str, lst: Iterable[str]) -> Iterator[str]:
        return filter(partial(_has_extension, ext=ext), lst)

    smry_candidates = filter_extension(
        r"unsmry|funsmry|s\d\d\d\d|a\d\d\d\d", smry_candidates
    )
    if specified_split:
        smry_candidates = filter_extension(r"s\d\d\d\d|a\d\d\d\d", smry_candidates)
    if specified_unified:
        smry_candidates = filter_extension("unsmry|funsmry", smry_candidates)
    if specified_formatted:
        smry_candidates = filter_extension("funsmry", smry_candidates)
    if specified_unformatted:
        smry_candidates = filter_extension("unsmry", smry_candidates)
    all_summary = natsorted(list(smry_candidates))
    summary = []
    pat = None
    for pat in ("unsmry", r"s\d\d\d\d", "funsmry", r"a\d\d\d\d"):
        summary = list(filter_extension(pat, all_summary))
        if summary:
            break

    if len(summary) != len(all_summary):
        warnings.warn(f"More than one type of summary file, found {all_summary}")
    if not summary:
        raise FileNotFoundError(f"Could not find any summary files matching {filepath}")

    if pat in ("unsmry", r"s\d\d\d\d"):
        spec_candidates = filter_extension("smspec", spec_candidates)
    else:
        spec_candidates = filter_extension("fsmspec", spec_candidates)

    spec = list(spec_candidates)
    if len(spec) > 1:
        warnings.warn(f"More than one type of summary spec file, found {spec}")

    if not spec:
        raise FileNotFoundError(f"Could not find any summary spec matching {filepath}")
    return summary, spec[-1]


def _stream_name(stream: IO[Any]) -> str:
    """
    Returns:
        The filename for an IO stream or 'unknown stream' if there is no filename
        attached to the stream (which is the case for eg. `StringIO` and `BytesIO`).
    """
    return getattr(stream, "name", "unknown stream")



"""
Implements a hypothesis strategy for unified summary files
(.SMSPEC and .UNSMRY)
See https://opm-project.org/?page_id=955
"""

from dataclasses import astuple, dataclass
from datetime import datetime, timedelta
from enum import Enum, unique
from typing import Any, Self, IO
from itertools import zip_longest
from os import PathLike

import hypothesis.strategies as st
import numpy as np
import resfo
from hypothesis import assume
from hypothesis.extra.numpy import from_dtype
from pydantic import PositiveInt, conint
from collections.abc import Iterator
import numpy.typing as npt

from ._egrid_generator import GrdeclKeyword
from resfo_utilities._summary_keys import SPECIAL_KEYWORDS
from resfo_utilities import make_summary_key

_inter_region_summary_variables = [
    "RGFR",
    "RGFR+",
    "RGFR-",
    "RGFT",
    "RGFT+",
    "RGFT-",
    "RGFTG",
    "RGFTL",
    "ROFR",
    "ROFR+",
    "ROFR-",
    "ROFT",
    "ROFT+",
    "ROFT-",
    "ROFTG",
    "ROFTL",
    "RWFR",
    "RWFR+",
    "RWFR-",
    "RWFT",
    "RWFT+",
    "RWFT-",
    "RCFT",
    "RSFT",
    "RNFT",
]


@st.composite
def _root_memnonic(draw: st.DrawFn) -> str:
    first_character = draw(st.sampled_from("ABFGRWCS"))
    if first_character == "A":
        second_character = draw(st.sampled_from("ALN"))
        third_character = draw(st.sampled_from("QL"))
        fourth_character = draw(st.sampled_from("RT"))
        return first_character + second_character + third_character + fourth_character
    else:
        second_character = draw(st.sampled_from("OWGVLPT"))
        third_character = draw(st.sampled_from("PIF"))
        fourth_character = draw(st.sampled_from("RT"))
        local = draw(st.sampled_from(["", "L"])) if first_character in "BCW" else ""
        return (
            local
            + first_character
            + second_character
            + third_character
            + fourth_character
        )


@st.composite
def summary_variables(draw: st.DrawFn) -> str:
    """Generator for valid summary variables.

    See the OPM Flow manual section 11.1.
    """
    kind = draw(
        st.sampled_from(
            [
                "special",
                "network",
                "exceptions",
                "directional",
                "up_or_down",
                "mnemonic",
                "segment",
                "well",
                "region2region",
                "mnemonic",
                "region",
            ]
        )
    )
    if kind == "special":
        return draw(st.sampled_from(SPECIAL_KEYWORDS))
    if kind == "exceptions":
        return draw(
            st.sampled_from(
                ["BAPI", "BOSAT", "BPR", "FAQR", "FPR", "FWCT", "WBHP", "WWCT", "ROFR"]
            )
        )
    elif kind == "directional":
        direction = draw(st.sampled_from("IJK"))
        return (
            draw(st.sampled_from(["FLOO", "VELG", "VELO", "FLOW", "VELW"])) + direction
        )
    elif kind == "up_or_down":
        dimension = draw(st.sampled_from("XYZRT"))
        direction = draw(st.sampled_from(["", "-"]))
        return draw(st.sampled_from(["GKR", "OKR", "WKR"])) + dimension + direction
    elif kind == "network":
        root = draw(_root_memnonic())
        return "N" + root
    elif kind == "segment":
        return draw(
            st.sampled_from(["SALQ", "SFR", "SGFR", "SGFRF", "SGFRS", "SGFTA", "SGFT"])
        )
    elif kind == "well":
        return draw(
            st.one_of(
                st.builds(lambda r: "W" + r, _root_memnonic()),
                st.sampled_from(
                    [
                        "WBHP",
                        "WBP5",
                        "WBP4",
                        "WBP9",
                        "WBP",
                        "WBHPH",
                        "WBHPT",
                        "WPIG",
                        "WPIL",
                        "WPIO",
                        "WPI5",
                    ]
                ),
            )
        )
    elif kind == "region2region":
        return draw(st.sampled_from(_inter_region_summary_variables))
    elif kind == "region":
        return draw(st.builds(lambda r: "R" + r, _root_memnonic()))
    else:
        return draw(_root_memnonic())


_unit_names = st.sampled_from(
    ["SM3/DAY", "BARSA", "SM3/SM3", "FRACTION", "DAYS", "HOURS", "SM3"]
)

_names = st.text(
    min_size=8,
    max_size=8,
    alphabet=st.characters(
        min_codepoint=65,
        max_codepoint=90,
    ),
)


@unique
class UnitSystem(Enum):
    """The unit system used for summary values."""

    METRIC = 1
    FIELD = 2
    LAB = 3

    def to_smry(self) -> int:
        return self.value


@unique
class Simulator(Enum):
    """The simulator used to generate the summary."""

    ECLIPSE_100 = 100
    ECLIPSE_300 = 300
    ECLIPSE_300_THERMAL = 500
    INTERSECT = 700
    FRONTSIM = 800

    def to_smry(self) -> int:
        return self.value


@dataclass
class SmspecIntehead(GrdeclKeyword):
    """The values in the INTEHEAD array"""

    unit: UnitSystem
    simulator: Simulator

    def to_smry(self) -> list[Any]:
        return [value.to_smry() for value in astuple(self)]


@dataclass
class Date:
    """The date given by the STARTDAT array"""

    day: conint(ge=1, le=31)  # type: ignore
    month: conint(ge=1, le=12)  # type: ignore
    year: conint(gt=1901, lt=2038)  # type: ignore
    hour: conint(ge=0, lt=24)  # type: ignore
    minutes: conint(ge=0, lt=60)  # type: ignore
    micro_seconds: conint(ge=0, lt=60000000)  # type: ignore

    def to_smry(self) -> tuple[int, ...]:
        return astuple(self)

    def to_datetime(self) -> datetime:
        return datetime(
            year=self.year,
            month=self.month,
            day=self.day,
            hour=self.hour,
            minute=self.minutes,
            second=self.micro_seconds // 10**6,
            microsecond=self.micro_seconds % 10**6,
        )

    @classmethod
    def from_datetime(cls, dt: datetime) -> Self:
        return cls(
            year=dt.year,
            month=dt.month,
            day=dt.day,
            hour=dt.hour,
            minutes=dt.minute,
            micro_seconds=dt.second * 10**6 + dt.microsecond,
        )


@dataclass
class Smspec:
    """The contents of the .SMSPEC file"""

    intehead: SmspecIntehead
    restart: str
    num_keywords: PositiveInt
    nx: PositiveInt
    ny: PositiveInt
    nz: PositiveInt
    restarted_from_step: PositiveInt
    keywords: list[str]
    well_names: list[str]
    region_numbers: list[int]
    units: list[str]
    start_date: Date
    lgrs: list[str] | None = None
    numlx: list[PositiveInt] | None = None
    numly: list[PositiveInt] | None = None
    numlz: list[PositiveInt] | None = None
    use_names: bool = False  # whether to use the alias NAMES for WGNAMES

    def to_smry(self) -> list[tuple[str, Any]]:
        # The restart field contains 9 strings of length 8 which
        # should contain the name of the file restarted from.
        # If shorter than 72 characters (most likely), the rest
        # are spaces. (opm manual table F.44, keyword name RESTART)
        restart = self.restart.ljust(72, " ")
        restart_list = [restart[i * 8 : i * 8 + 8] for i in range(9)]
        return (
            [
                ("INTEHEAD", np.array(self.intehead.to_smry(), dtype=np.int32)),
                ("RESTART ", restart_list),
                (
                    "DIMENS  ",
                    np.array(
                        [
                            self.num_keywords,
                            self.nx,
                            self.ny,
                            self.nz,
                            0,
                            self.restarted_from_step,
                        ],
                        dtype=np.int32,
                    ),
                ),
                ("KEYWORDS", [kw.ljust(8) for kw in self.keywords]),
                (
                    ("NAMES   ", self.well_names)
                    if self.use_names
                    else ("WGNAMES ", self.well_names)
                ),
                ("NUMS    ", np.array(self.region_numbers, dtype=np.int32)),
                ("UNITS   ", self.units),
                ("STARTDAT", np.array(self.start_date.to_smry(), dtype=np.int32)),
            ]
            + ([("LGRS    ", self.lgrs)] if self.lgrs is not None else [])
            + ([("NUMLX   ", self.numlx)] if self.numlx is not None else [])
            + ([("NUMLY   ", self.numly)] if self.numly is not None else [])
            + ([("NUMLZ   ", self.numlz)] if self.numlz is not None else [])
        )

    def summary_keys(self) -> Iterator[str]:
        """All summary keys in the SMSPEC file."""

        def optional(maybe_list: list[Any] | None) -> list[Any]:
            if maybe_list is None:
                return []
            else:
                return maybe_list

        for var, num, name, lgr, lx, ly, lz in zip_longest(
            self.keywords,
            self.region_numbers,
            self.well_names,
            optional(self.lgrs),
            optional(self.numlx),
            optional(self.numly),
            optional(self.numlz),
        ):
            yield make_summary_key(var, num, name, self.nx, self.ny, lgr, lx, ly, lz)

    def to_file(
        self,
        filelike: str | PathLike[str] | IO[Any],
        file_format: resfo.Format = resfo.Format.UNFORMATTED,
    ) -> None:
        """Writes the Smspec to a file"""
        resfo.write(filelike, self.to_smry(), file_format)


_small_ints = from_dtype(np.dtype(np.int32), min_value=1, max_value=10)


@st.composite
def smspecs(
    draw: st.DrawFn,
    sum_keys: st.SearchStrategy[list[str]],
    start_date: st.SearchStrategy[Date],
    use_days: st.SearchStrategy[bool] | None = None,
    well_names: st.SearchStrategy[str] | None = None,
    lgr_names: st.SearchStrategy[str] | None = None,
    restart_names: st.SearchStrategy[str] | None = None,
) -> Smspec:
    """Hypothesis strategy for ``Smspec`` s."""
    use_days = st.booleans() if use_days is None else use_days
    use_locals = draw(st.booleans())
    sum_keys_ = draw(sum_keys)
    if any(sk.startswith("L") for sk in sum_keys_):
        use_locals = True
    n = len(sum_keys_) + 1
    nx = draw(_small_ints)
    ny = draw(_small_ints)
    nz = draw(_small_ints)
    keywords = ["TIME    "] + sum_keys_
    if draw(use_days):
        units = [
            "DAYS    ",
            *draw(st.lists(_unit_names, min_size=n - 1, max_size=n - 1)),
        ]
    else:
        units = [
            "HOURS   ",
            *draw(st.lists(_unit_names, min_size=n - 1, max_size=n - 1)),
        ]
    well_names_ = [
        ":+:+:+:+",
        *draw(st.lists(well_names or _names, min_size=n - 1, max_size=n - 1)),
    ]
    if use_locals:  # use local
        lgrs = draw(st.lists(lgr_names or _names, min_size=n, max_size=n))
        numlx = draw(st.lists(_small_ints, min_size=n, max_size=n))
        numly = draw(st.lists(_small_ints, min_size=n, max_size=n))
        numlz = draw(st.lists(_small_ints, min_size=n, max_size=n))
    else:
        lgrs = None
        numlx = None
        numly = None
        numlz = None
    region_numbers = [-32676] + draw(
        st.lists(
            from_dtype(np.dtype(np.int32), min_value=1, max_value=nx * ny * nz),
            min_size=len(sum_keys_),
            max_size=len(sum_keys_),
        )
    )
    return draw(
        st.builds(
            Smspec,
            nx=st.just(nx),
            ny=st.just(ny),
            nz=st.just(nz),
            # restarted_from_step is hardcoded to 0 because
            # of a bug in enkf_obs where it assumes that
            # ecl_sum_get_first_report_step is always 1
            restarted_from_step=st.just(0),
            num_keywords=st.just(n),
            restart=restart_names or _names,
            keywords=st.just(keywords),
            well_names=st.just(well_names_),
            lgrs=st.just(lgrs),
            numlx=st.just(numlx),
            numly=st.just(numly),
            numlz=st.just(numlz),
            region_numbers=st.just(region_numbers),
            units=st.just(units),
            start_date=start_date,
            use_names=st.booleans(),
        )
    )


@dataclass
class SummaryMiniStep:
    """One ministep section in a summary file."""

    mini_step: int
    params: list[float]

    def to_smry(self) -> list[tuple[str, npt.NDArray[Any]]]:
        return [
            ("MINISTEP", np.array([self.mini_step], dtype=np.int32)),
            ("PARAMS  ", np.array(self.params, dtype=np.float32)),
        ]


@dataclass
class SummaryStep:
    """One step section in a summary file.

    Is just one SEQHDR followed by one or more ministeps.
    """

    seqnum: int
    ministeps: list[SummaryMiniStep]

    def to_smry(self) -> list[tuple[str, npt.NDArray[Any]]]:
        return [("SEQHDR  ", np.array([self.seqnum], dtype=np.int32))] + [
            i for ms in self.ministeps for i in ms.to_smry()
        ]


@dataclass
class Unsmry:
    """The contents of a unified summary file."""

    steps: list[SummaryStep]

    def to_smry(self) -> list[tuple[str, npt.NDArray[Any]]]:
        a = [i for step in self.steps for i in step.to_smry()]
        return a

    def to_file(
        self,
        filelike: str | PathLike[str] | IO[Any],
        file_format: resfo.Format = resfo.Format.UNFORMATTED,
    ) -> None:
        resfo.write(filelike, self.to_smry(), file_format)


_positive_floats = from_dtype(
    np.dtype(np.float32),
    min_value=np.float32(0.1),  # type: ignore
    max_value=np.float32(1e19),  # type: ignore
    allow_nan=False,
    allow_infinity=False,
)


_start_dates = st.datetimes(
    min_value=datetime.strptime("1969-1-1", "%Y-%m-%d"),
    max_value=datetime.strptime("2100-1-1", "%Y-%m-%d"),
)

_time_delta_lists = st.lists(
    st.floats(
        min_value=0.1,
        max_value=2500.0,  # in days ~= 6.8 years
        allow_nan=False,
        allow_infinity=False,
    ),
    min_size=2,
    max_size=100,
    unique=True,
)

_summary_keys = st.lists(summary_variables(), min_size=1)


@st.composite
def summaries(
    draw: st.DrawFn,
    start_date: st.SearchStrategy[datetime] = _start_dates,
    time_deltas: st.SearchStrategy[list[float]] = _time_delta_lists,
    summary_keys: st.SearchStrategy[list[str]] = _summary_keys,
    use_days: st.SearchStrategy[bool] | None = None,
    report_step_only: bool = False,
) -> tuple[Smspec, Unsmry]:
    """Generator of a smspec and unsmry pair.

    The strategy ensures that the files matches in number of keywords.
    """
    sum_keys = draw(summary_keys)
    first_date = draw(start_date)
    days = draw(use_days if use_days is not None else st.booleans())
    smspec = draw(
        smspecs(
            sum_keys=st.just(sum_keys),
            start_date=st.just(Date.from_datetime(first_date)),
            use_days=st.just(days),
        )
    )
    # The smspec should be unique up to summary_keys.
    # This just mimics the behavior of simulators.
    assume(len(set(smspec.summary_keys())) == len(smspec.keywords))
    dates = [0.0] + draw(time_deltas)
    try:
        if days:
            _ = first_date + timedelta(days=max(dates))
        else:
            _ = first_date + timedelta(hours=max(dates))
    except (ValueError, OverflowError):  # datetime has a max year
        assume(False)

    ds = sorted(dates, reverse=True)
    steps = []
    i = 0
    j = 0
    while len(ds) > 0:
        minis = []
        max_val = 1 if report_step_only else len(ds)
        for _ in range(draw(st.integers(min_value=1, max_value=max_val))):
            minis.append(
                SummaryMiniStep(
                    i,
                    [
                        ds.pop(),
                        *draw(
                            st.lists(
                                _positive_floats,
                                min_size=len(sum_keys),
                                max_size=len(sum_keys),
                            )
                        ),
                    ],
                )
            )
            i += 1
        steps.append(SummaryStep(j, minis))
        j += 1
    return smspec, Unsmry(steps)


__all__ = [
    "summary_variables",
    "UnitSystem",
    "Simulator",
    "SmspecIntehead",
    "Date",
    "Smspec",
    "smspecs",
    "SummaryMiniStep",
    "SummaryStep",
    "Unsmry",
    "summaries",
]



from resfo_utilities.testing import GlobalGrid
from hypothesis import given
import hypothesis.strategies as st


@given(st.builds(GlobalGrid))
def test_that_global_grid_eq_is_reflexive(global_grid):
    assert global_grid == global_grid


@given(*([st.builds(GlobalGrid)] * 3))
def test_that_global_grid_eq_is_transitive(a, b, c):
    if a == b and b == c:
        assert a == c



import pytest
import shutil
import os


@pytest.fixture
def simulator_cmd():
    simulator_cmd = os.environ.get("TEST_SIMULATOR_CMD") or shutil.which("flow")
    if not simulator_cmd:
        pytest.skip(
            reason="Did not find OPM flow simulator"
            " and no simulator set in TEST_SIMULATOR_CMD"
        )
    return simulator_cmd.split()



import hypothesis.strategies as st
import pytest
from hypothesis import given

from resfo_utilities import SummaryKeyType, make_summary_key, is_rate
from resfo_utilities.testing._summary_generator import _inter_region_summary_variables

DUMMY_NAME = ":+:+:+:+"


@pytest.mark.parametrize("keyword", ["AAQR", "AAQT"])
def test_aquifer_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.AQUIFER


@pytest.mark.parametrize("keyword", ["BOSAT"])
def test_block_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.BLOCK


@pytest.mark.parametrize("keyword", ["LBOSAT"])
def test_local_block_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.LOCAL_BLOCK


@pytest.mark.parametrize("keyword", ["CGORL"])
def test_completion_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.COMPLETION


@pytest.mark.parametrize("keyword", ["LCGORL"])
def test_local_completion_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.LOCAL_COMPLETION


@pytest.mark.parametrize("keyword", ["FGOR", "FOPR"])
def test_field_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.FIELD


@pytest.mark.parametrize("keyword", ["GGFT", "GOPR"])
def test_group_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.GROUP


@pytest.mark.parametrize("keyword", ["NOPR", "NGPR"])
def test_network_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.NETWORK


@pytest.mark.parametrize("keyword", _inter_region_summary_variables)
def test_inter_region_summary_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.INTER_REGION


@pytest.mark.parametrize("keyword", ["RORFR", "RPR", "ROPT"])
def test_region_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.REGION


@pytest.mark.parametrize("keyword", ["SOPR"])
def test_segment_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.SEGMENT


@pytest.mark.parametrize("keyword", ["WOPR"])
def test_well_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.WELL


@pytest.mark.parametrize("keyword", ["LWOPR"])
def test_local_well_variables_are_recognized(keyword):
    assert SummaryKeyType.from_variable(keyword) == SummaryKeyType.LOCAL_WELL


positives = st.integers(min_value=0)
valid_names = st.text(min_size=1).filter(lambda name: name != DUMMY_NAME)


@given(st.integers(), valid_names, positives, positives)
@pytest.mark.parametrize("keyword", ["FOPR", "NEWTON"])
def test_summary_key_format_of_field_and_misc_is_identity(
    keyword, number, name, nx, ny
):
    assert make_summary_key(keyword, number, name, nx, ny) == keyword


@given(st.integers(), valid_names, positives, positives)
def test_network_variable_keys_has_keyword_as_summary_key(number, name, nx, ny):
    assert make_summary_key("NOPR", number, name, nx, ny) == f"NOPR:{name}"


@given(
    positives,
    valid_names,
    positives,
    positives,
)
@pytest.mark.parametrize("keyword", ["GOPR", "WOPR"])
def test_group_and_well_have_named_format(keyword, number, name, nx, ny):
    assert make_summary_key(keyword, number, name, nx, ny) == f"{keyword}:{name}"


@given(
    st.sampled_from(_inter_region_summary_variables),
    valid_names,
    positives,
    positives,
)
def test_inter_region_summary_format_contains_in_and_out_regions(keyword, name, nx, ny):
    number = 3014660
    assert make_summary_key(keyword, number, name, nx, ny) == f"{keyword}:4-82"


@given(name=valid_names)
@pytest.mark.parametrize("keyword", ["BOPR", "BOSAT"])
@pytest.mark.parametrize(
    "nx,ny,number,indices",
    [
        (1, 1, 1, "1,1,1"),
        (2, 1, 2, "2,1,1"),
        (1, 2, 2, "1,2,1"),
        (3, 2, 3, "3,1,1"),
        (3, 2, 9, "3,1,2"),
    ],
)
def test_block_summary_format_have_cell_index(keyword, number, indices, name, nx, ny):
    assert make_summary_key(keyword, number, name, nx, ny) == f"{keyword}:{indices}"


@given(name=valid_names)
@pytest.mark.parametrize("keyword", ["COPR"])
@pytest.mark.parametrize(
    "nx,ny,number,indices",
    [
        (1, 1, 1, "1,1,1"),
        (2, 1, 2, "2,1,1"),
        (1, 2, 2, "1,2,1"),
        (3, 2, 3, "3,1,1"),
        (3, 2, 9, "3,1,2"),
    ],
)
def test_completion_summary_format_have_cell_index_and_name(
    keyword, number, indices, name, nx, ny
):
    assert (
        make_summary_key(keyword, number, name, nx, ny) == f"{keyword}:{name}:{indices}"
    )


@pytest.mark.parametrize("keyword", ["LBWPR"])
@pytest.mark.parametrize(
    "li,lj,lk,lgr_name,indices",
    [
        (1, 1, 1, "LGRNAME", "1,1,1"),
        (2, 1, 1, "LGRNAME", "2,1,1"),
        (1, 2, 1, "LGRNAME", "1,2,1"),
        (3, 1, 1, "LGRNAME", "3,1,1"),
        (3, 1, 2, "LGRNAME", "3,1,2"),
    ],
)
def test_local_block_summary_format_have_cell_index_and_name(
    keyword, lgr_name, indices, li, lj, lk
):
    assert (
        make_summary_key(keyword, li=li, lj=lj, lk=lk, lgr_name=lgr_name)
        == f"{keyword}:{lgr_name}:{indices}"
    )


@given(name=valid_names, lgr_name=st.text())
@pytest.mark.parametrize("keyword", ["LCOPR"])
@pytest.mark.parametrize(
    "li,lj,lk,indices",
    [
        (1, 1, 1, "1,1,1"),
        (2, 1, 1, "2,1,1"),
        (1, 2, 1, "1,2,1"),
        (3, 1, 1, "3,1,1"),
        (3, 1, 2, "3,1,2"),
    ],
)
def test_local_completion_summary_format_have_cell_index_and_name(
    keyword, name, lgr_name, indices, li, lj, lk
):
    assert (
        make_summary_key(keyword, name=name, li=li, lj=lj, lk=lk, lgr_name=lgr_name)
        == f"{keyword}:{lgr_name}:{name}:{indices}"
    )


@given(name=valid_names, lgr_name=st.text())
@pytest.mark.parametrize("keyword", ["LWWPR"])
def test_local_well_summary_format_have_cell_index_and_name(keyword, name, lgr_name):
    assert (
        make_summary_key(keyword, name=name, lgr_name=lgr_name)
        == f"{keyword}:{lgr_name}:{name}"
    )


def nonempty_string_without_whitespace():
    return st.text(
        st.characters(whitelist_categories=("Lu", "Ll", "Nd", "P")), min_size=1
    )


@given(key=nonempty_string_without_whitespace())
def test_is_rate_does_not_raise_error(key):
    is_rate_bool = is_rate(key)
    assert isinstance(is_rate_bool, bool)


@pytest.mark.parametrize(
    "opr_summary_variable", ["FOPR", "FOPR", "GOPR", "WOPR", "COPR", "ROPR", "LCOPR"]
)
def test_that_oil_production_rate_is_a_rate(opr_summary_variable):
    assert is_rate(opr_summary_variable)


@pytest.mark.parametrize(
    "opt_summary_variable", ["FOPT", "FOPT", "GOPT", "WOPT", "COPT", "ROPT", "LCOPT"]
)
def test_that_oil_production_total_is_not_a_rate(opt_summary_variable):
    assert not is_rate(opt_summary_variable)


@pytest.mark.parametrize(
    "gpr_summary_variable", ["FGPR", "FGPR", "GGPR", "WGPR", "CGPR", "RGPR", "LCGPR"]
)
def test_that_gas_production_rate_is_a_rate(gpr_summary_variable):
    assert is_rate(gpr_summary_variable)


@pytest.mark.parametrize(
    "opt_summary_variable", ["FGPT", "FGPT", "GGPT", "WGPT", "CGPT", "RGPT", "LCGPT"]
)
def test_that_gas_production_total_is_not_a_rate(opt_summary_variable):
    assert not is_rate(opt_summary_variable)


@pytest.mark.parametrize(
    "gir_summary_variable", ["FGIR", "FGIR", "GGIR", "WGIR", "CGIR", "RGIR", "LCGIR"]
)
def test_that_gas_injection_rate_is_a_rate(gir_summary_variable):
    assert is_rate(gir_summary_variable)


@pytest.mark.parametrize(
    "git_summary_variable", ["FGIT", "FGIT", "GGIT", "WGIT", "CGIT", "RGIT", "LCGIT"]
)
def test_that_gas_injection_total_is_not_a_rate(git_summary_variable):
    assert not is_rate(git_summary_variable)


@pytest.mark.parametrize(
    "wpr_summary_variable", ["FWPR", "FWPR", "GWPR", "WWPR", "CWPR", "RWPR", "LCWPR"]
)
def test_that_water_production_rate_is_a_rate(wpr_summary_variable):
    assert is_rate(wpr_summary_variable)


@pytest.mark.parametrize(
    "wpt_summary_variable", ["FWPT", "FGPT", "GGPT", "WGPT", "CGPT", "RGPT", "LCGPT"]
)
def test_that_water_production_total_is_not_a_rate(wpt_summary_variable):
    assert not is_rate(wpt_summary_variable)


@pytest.mark.parametrize(
    "wir_summary_variable", ["FWIR", "FWIR", "GWIR", "WWIR", "CWIR", "RWIR", "LCWIR"]
)
def test_that_water_injection_rate_is_a_rate(wir_summary_variable):
    assert is_rate(wir_summary_variable)


@pytest.mark.parametrize(
    "wit_summary_variable", ["FWIT", "FWIT", "GWIT", "WWIT", "CWIT", "RWIT", "LCWIT"]
)
def test_that_water_injection_total_is_not_a_rate(wit_summary_variable):
    assert not is_rate(wit_summary_variable)


@pytest.mark.parametrize("pr_summary_variable", ["FPR", "RPR", "BPR"])
def test_that_oil_phase_pressure_is_not_a_rate(pr_summary_variable):
    assert not is_rate(pr_summary_variable)


@pytest.mark.parametrize("lpr_summary_variable", ["FLPR", "GLPR", "WLPR", "CLPR"])
def test_that_liquid_production_rate_is_a_rate(lpr_summary_variable):
    assert is_rate(lpr_summary_variable)


@pytest.mark.parametrize("lpt_summary_variable", ["FLPT", "GLPT", "WLPT", "CLPT"])
def test_that_liquid_production_total_is_not_a_rate(lpt_summary_variable):
    assert not is_rate(lpt_summary_variable)


@pytest.mark.parametrize("mwpr_summary_variable", ["FMWPR", "GMWPR"])
def test_that_number_of_production_wells_flowing_is_not_a_rate(mwpr_summary_variable):
    assert not is_rate(mwpr_summary_variable)


@pytest.mark.parametrize(
    "gor_summary_variable", ["FGOR", "GGOR", "WGOR", "CGOR", "CGORL", "SGOR"]
)
def test_that_gas_oil_ratio_is_a_rate(gor_summary_variable):
    assert is_rate(gor_summary_variable)


@pytest.mark.parametrize(
    "ogr_summary_variable", ["FOGR", "GOGR", "WOGR", "COGR", "COGRL", "SOGR"]
)
def test_that_oil_gas_ratio_is_a_rate(ogr_summary_variable):
    assert is_rate(ogr_summary_variable)


@pytest.mark.parametrize(
    "wgr_summary_variable", ["FWGR", "GWGR", "WWGR", "CWGR", "CWGRL", "SWGR"]
)
def test_that_water_gas_ratio_is_a_rate(wgr_summary_variable):
    assert is_rate(wgr_summary_variable)


@pytest.mark.parametrize(
    "glr_summary_variable", ["FGLR", "GGLR", "WGLR", "CGLR", "CGLRL"]
)
def test_that_gas_liquid_ratio_is_a_rate(glr_summary_variable):
    assert is_rate(glr_summary_variable)


@pytest.mark.parametrize(
    "wct_summary_variable", ["FWCT", "GWCT", "WWCT", "CWCT", "CWCTL", "SWCT"]
)
def test_that_water_cut_is_a_rate(wct_summary_variable):
    assert is_rate(wct_summary_variable)
